import os
import shutil
import tempfile
import threading
from importlib import resources
from pathlib import Path

from py4j.java_gateway import JavaObject
from pyspark import SparkContext


# Environment variables for the jar path. It is only used for integration tests because in local integration tests the
# file is not accessible via the resources.path. We need to access it via the Bazel runfiles.
_UDF_JAR_PATH = os.environ.get("UDF_JAR_PATH_FOR_INTEGRATION_TEST")


class SingletonResourceJarClassLoader:
    """Wraps a Java ClassLoader with its classpath set to a JAR.
    Additionally, upon initialization the JAR file is added to the SparkContext JARs, which will cause it to be added
    to the class path of executors. This is so that e.g. UDFs loaded from this class loader on the driver will be able
    to find their classes on the executors during deserialization.
    """

    _instance = None
    _init_lock = threading.Lock()

    def __new__(cls, jar: Path) -> None:
        with cls._init_lock:
            if not cls._instance:
                cls._instance = super(SingletonResourceJarClassLoader, cls).__new__(cls)
                cls._instance._initialize(jar)
            return cls._instance

    def _initialize(self, jar: Path) -> None:
        self._jar = tempfile.NamedTemporaryFile()
        shutil.copyfile(jar, self._jar.name)
        self._lock = threading.Lock()
        self._loader = None

    def get_loader(self, sc: SparkContext) -> JavaObject:
        """Lazily initializes and returns the ClassLoader object."""
        with self._lock:
            if not self._loader:
                gateway = sc._gateway
                # This tells Spark to ship the JAR to the executors and add it to the class path.
                sc._jsc.addJar(self._jar.name)
                jvm = gateway.jvm
                URL = jvm.java.net.URL
                URLClassLoader = jvm.java.net.URLClassLoader
                url_array = gateway.new_array(URL, 1)
                url_array[0] = URL(f"file://{self._jar.name}")
                self._loader = URLClassLoader(url_array)
            return self._loader

    def load_class(self, sc: SparkContext, class_name: str) -> JavaObject:
        """Loads and initializes the class"""
        return sc._gateway.jvm.java.lang.Class.forName(class_name, True, self.get_loader(sc))


def get_or_create_udf_jar_class_loader():
    if _UDF_JAR_PATH:
        return SingletonResourceJarClassLoader(Path(_UDF_JAR_PATH))
    else:
        with resources.path("tecton_spark.jars", "tecton-udfs-spark-3.jar") as p:
            return SingletonResourceJarClassLoader(p)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/args/fco_args__client.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from tecton_proto.args import entity__client_pb2 as tecton__proto_dot_args_dot_entity____client__pb2
from tecton_proto.args import feature_service__client_pb2 as tecton__proto_dot_args_dot_feature__service____client__pb2
from tecton_proto.args import feature_view__client_pb2 as tecton__proto_dot_args_dot_feature__view____client__pb2
from tecton_proto.args import transformation__client_pb2 as tecton__proto_dot_args_dot_transformation____client__pb2
from tecton_proto.args import virtual_data_source__client_pb2 as tecton__proto_dot_args_dot_virtual__data__source____client__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n(tecton_proto/args/fco_args__client.proto\x12\x11tecton_proto.args\x1a&tecton_proto/args/entity__client.proto\x1a/tecton_proto/args/feature_service__client.proto\x1a,tecton_proto/args/feature_view__client.proto\x1a.tecton_proto/args/transformation__client.proto\x1a\x33tecton_proto/args/virtual_data_source__client.proto\"\x9e\x03\n\x07\x46\x63oArgs\x12Z\n\x13virtual_data_source\x18\x01 \x01(\x0b\x32(.tecton_proto.args.VirtualDataSourceArgsH\x00R\x11virtualDataSource\x12\x37\n\x06\x65ntity\x18\x02 \x01(\x0b\x32\x1d.tecton_proto.args.EntityArgsH\x00R\x06\x65ntity\x12G\n\x0c\x66\x65\x61ture_view\x18\x06 \x01(\x0b\x32\".tecton_proto.args.FeatureViewArgsH\x00R\x0b\x66\x65\x61tureView\x12P\n\x0f\x66\x65\x61ture_service\x18\x04 \x01(\x0b\x32%.tecton_proto.args.FeatureServiceArgsH\x00R\x0e\x66\x65\x61tureService\x12O\n\x0etransformation\x18\x07 \x01(\x0b\x32%.tecton_proto.args.TransformationArgsH\x00R\x0etransformationB\x06\n\x04\x61rgsJ\x04\x08\x03\x10\x04J\x04\x08\x05\x10\x06\x42\x13\n\x0f\x63om.tecton.argsP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.args.fco_args__client_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\017com.tecton.argsP\001'
  _FCOARGS._serialized_start=300
  _FCOARGS._serialized_end=714
# @@protoc_insertion_point(module_scope)



from typing import Any
from typing import Iterable
from typing import List
from typing import Optional
from typing import Tuple
from typing import TypeVar

import attrs
import pendulum
import typeguard
from attrs import validators
from google import protobuf
from google.protobuf import duration_pb2

from tecton_core import time_utils


# An attrs metadata key use to indicate that a spec field is allowed and expected to diverge for the same object
# definition between local and remote objects. This metadata used in testing and to codify fields where divergence is
# permitted.
LOCAL_REMOTE_DIVERGENCE_ALLOWED = "local_remote_divergence_allowed"


@typeguard.typechecked
def get_field_or_none(proto: protobuf.message.Message, field: str) -> Any:  # noqa: ANN401
    """Get the proto message's field. Return None if the field is not set (instead of the default value).

    When filling specs, `None` should be typically filled when the underlying proto field is unset - except in cases
    where the proto default (e.g. 0) is the desired for the Python value.
    """
    if proto.HasField(field):
        return getattr(proto, field)
    else:
        return None


@typeguard.typechecked
def get_non_default_duration_field_or_none(proto: protobuf.message.Message, field: str) -> Optional[pendulum.Duration]:
    """Get the proto message's duration-type field. If it equals the default value or does not exist, return None."""
    if proto.HasField(field):
        proto_duration = getattr(proto, field)
        if proto_duration != duration_pb2.Duration():
            return time_utils.proto_to_duration(proto_duration)

    return None


@typeguard.typechecked
def get_duration_field_or_none(proto: protobuf.message.Message, field: str) -> Optional[pendulum.Duration]:
    """Get the proto message's duration-type field. Return None if the field is not set (instead of the default value)."""
    if proto.HasField(field):
        proto_duration = getattr(proto, field)
        return time_utils.proto_to_duration(proto_duration)
    else:
        return None


@typeguard.typechecked
def get_timestamp_field_or_none(proto: protobuf.message.Message, field: str) -> Optional[pendulum.DateTime]:
    """Get the proto message's timestamp-type field. Return None if the field is not set (instead of the default value)."""
    if proto.HasField(field):
        proto_timestamp = getattr(proto, field)
        return pendulum.instance(proto_timestamp.ToDatetime())
    else:
        return None


T = TypeVar("T")


@typeguard.typechecked
def get_tuple_from_repeated_field(repeated_field: Iterable[T]) -> Tuple[T, ...]:
    return tuple(value for value in repeated_field)


def type_validator(instance: Any, attribute: attrs.Attribute, value: Any) -> None:  # noqa: ANN401
    """An attrs validator that asserts that an attribute matches its declared type."""
    assert attribute.type is not None, "Type annotations are required."

    # It would be better to assert on the affirmative (e.g. `assert inspect.is_class(attribute.type)`), but generic
    # types (e.g. `Optional[str]`) are not classes and do not have a documented way to identify them.
    assert not isinstance(
        attribute.type, str
    ), f"Found unresolved type annotation `{attribute.type}` for attribute `{attribute.name}`. Do not use forward-declared types with the frozen_strict decorator."
    try:
        typeguard.check_type(attribute.name, value, attribute.type)
    except TypeError as e:
        # Needed because generic types (e.g. `Tuple[str, ...]`) do not have a __name__ attribute.
        attribute_type = attribute.type.__name__ if hasattr(attribute.type, "__name__") else str(attribute.type)
        msg = f"{instance.__class__.__name__} initialized with invalid type for attribute `{attribute.name}`. Expected type: `{attribute_type}`. Provided type: `{type(value).__name__}`."
        raise TypeError(msg) from e


def add_strict_type_validation(cls: type, fields: List[attrs.Attribute]) -> List[attrs.Attribute]:
    """An attrs field transformer that add type assertions to all fields."""
    new_fields = []
    for field in fields:
        if field.validator:
            new_validator = validators.and_(type_validator, field.validator)
        else:
            new_validator = type_validator
        new_fields.append(field.evolve(validator=new_validator))
    return new_fields


def frozen_strict(cls):
    """A decorator used to define a frozen attrs class where all attribute type annotations are enforced at runtime."""

    # Use slots=False because the slots optimization causes issues with serialization (needed for pyspark UDFs run on
    # worker nodes) when combined with attrs inheritance, i.e. attrs classes inheriting from other attrs classes.
    return attrs.frozen(cls, field_transformer=add_strict_type_validation, slots=False)


@frozen_strict
class JoinKeyMappingSpec:
    spine_column_name: str
    feature_view_column_name: str



import io
import json
import logging
from typing import Optional
from typing import Union

import attrs
import pandas
import pyarrow
import pyarrow.parquet as pq
from google.api_core import client_info
from google.cloud import bigquery
from google.cloud.bigquery import _pandas_helpers

from tecton_core import _gen_version
from tecton_core import conf
from tecton_core.errors import TectonInternalError
from tecton_core.query.dialect import Dialect
from tecton_core.query.node_interface import NodeRef
from tecton_core.query.nodes import DataSourceScanNode
from tecton_core.query.query_tree_compute import ComputeMonitor
from tecton_core.query.query_tree_compute import SQLCompute
from tecton_core.query.sql_compat import CompatFunctions
from tecton_core.schema import Schema
from tecton_core.schema_validation import cast_batch
from tecton_core.schema_validation import tecton_schema_to_arrow_schema
from tecton_core.secrets import SecretResolver
from tecton_core.specs import BigquerySourceSpec


@attrs.define
class BigqueryCompute(SQLCompute):
    is_debug: bool = attrs.field(init=False)

    def __attrs_post_init__(self):
        self.is_debug = conf.get_bool("DUCKDB_DEBUG")

    def get_dialect(self) -> Dialect:
        return Dialect.BIGQUERY

    def run_sql(
        self,
        sql_string: str,
        return_dataframe: bool = False,
        expected_output_schema: Optional[Schema] = None,
        monitor: Optional[ComputeMonitor] = None,
    ) -> Optional[pyarrow.RecordBatchReader]:
        # Although `sqlparse.format` may be helpful, it has been slow at times. BQ error logs are generally good.
        if self.is_debug:
            logging.warning(f"BigQuery QT: run SQL:\n{sql_string}")

        if monitor:
            monitor.set_query(sql_string)

        client = BigquerySessionManager.get_client()
        query_job_config = BigquerySessionManager.get_query_job_config()
        query_job = client.query(sql_string, job_config=query_job_config).result()

        if not return_dataframe:
            return

        if expected_output_schema:
            output_schema = tecton_schema_to_arrow_schema(expected_output_schema)
        else:
            output_schema = _pandas_helpers.bq_to_arrow_schema(query_job.schema)

        def batch_iterator():
            for batch in query_job.to_arrow_iterable():
                yield cast_batch(batch, output_schema)

        return pyarrow.RecordBatchReader.from_batches(output_schema, batch_iterator())

    def register_temp_table(
        self, table_name: str, table_or_reader: Union[pyarrow.Table, pyarrow.RecordBatchReader]
    ) -> None:
        if self.is_debug:
            logging.warning(f"BigQuery QT: Registering temp table: _SESSION.{table_name}")

        client = BigquerySessionManager.get_client()
        load_job_config = BigquerySessionManager.get_load_job_config()
        load_job_config.source_format = bigquery.SourceFormat.PARQUET
        load_job_config.autodetect = True

        # Note: load_table_from_dataframe converts pandas to parquet and calls load_table_from_file
        # So we write the arrow batches directly to parquet without converting to a pandas dataframe
        with io.BytesIO() as stream:
            if isinstance(table_or_reader, pyarrow.RecordBatchReader):
                writer = None
                for batch in table_or_reader:
                    if writer is None:
                        writer = pq.ParquetWriter(stream, batch.schema)
                    writer.write_batch(batch)
                if writer:
                    writer.close()
            else:
                pq.write_table(table_or_reader, stream)
            stream.seek(0)
            load_job = client.load_table_from_file(stream, f"_SESSION.{table_name}", job_config=load_job_config)
            load_job.result()

    def register_temp_table_from_pandas(self, table_name: str, pandas_df: pandas.DataFrame) -> None:
        if self.is_debug:
            logging.warning(f"BigQuery QT: Registering temp table from pandas: _SESSION.{table_name}")
        client = BigquerySessionManager.get_client()
        load_job_config = BigquerySessionManager.get_load_job_config()
        load_job = client.load_table_from_dataframe(pandas_df, f"_SESSION.{table_name}", job_config=load_job_config)
        load_job.result()

    def register_temp_table_from_data_source(
        self,
        table_name: str,
        ds_node: DataSourceScanNode,
        secret_resolver: Optional[SecretResolver] = None,
        monitor: Optional[ComputeMonitor] = None,
    ) -> None:
        ds_spec = ds_node.ds.batch_source
        assert isinstance(ds_spec, BigquerySourceSpec), "BigQuery compute supports only BigQuery data sources"

        query_str = ds_node.with_dialect(Dialect.BIGQUERY).to_sql()

        if monitor:
            monitor.set_query(query_str)

        # TODO: for cases where ds compute != pipeline compute, it maybe better to register duckdb table
        # this avoids the need for bq create perms if customer just wants to limit tecton to read bq tables
        run_sql = f"CREATE TEMP TABLE {table_name} AS ({query_str})"
        client = BigquerySessionManager.get_client(ds_spec, secret_resolver)
        query_job_config = BigquerySessionManager.get_query_job_config()
        client.query(run_sql, job_config=query_job_config).result()

    def load_table(self, table_name: str, expected_output_schema: Optional[Schema] = None) -> pyarrow.RecordBatchReader:
        return self.run_sql(
            CompatFunctions.for_dialect(Dialect.BIGQUERY).query().from_(table_name).select("*").get_sql(),
            return_dataframe=True,
            expected_output_schema=expected_output_schema,
        )

    def cleanup_temp_tables(self):
        if self.is_debug:
            logging.warning("Cleaning up BigQuery session")
        BigquerySessionManager.abort_session()

    def run_odfv(
        self, qt_node: NodeRef, input_df: pandas.DataFrame, monitor: Optional[ComputeMonitor] = None
    ) -> pandas.DataFrame:
        raise NotImplementedError


class BigquerySessionManager:
    _session_id = None
    _session_client = None
    _user_agent = f"tecton/sdk/{_gen_version.VERSION}"

    @classmethod
    def _initiate_session(cls):
        if cls._session_id is None:
            job = cls._session_client.query(
                "SELECT 1;",
                job_config=bigquery.QueryJobConfig(create_session=True),
            )
            cls._session_id = job.session_info.session_id
            job.result()

    @classmethod
    def abort_session(cls):
        if cls._session_id:
            job = cls._session_client.query(
                "CALL BQ.ABORT_SESSION();",
                job_config=bigquery.QueryJobConfig(
                    create_session=False,
                    connection_properties=[bigquery.query.ConnectionProperty(key="session_id", value=cls._session_id)],
                ),
            )
            job.result()
            cls._session_id = None
            cls._session_client = None

    @classmethod
    def get_query_job_config(cls) -> "bigquery.QueryJobConfig":
        return bigquery.QueryJobConfig(
            create_session=False,
            connection_properties=[bigquery.query.ConnectionProperty(key="session_id", value=cls._session_id)],
        )

    @classmethod
    def get_load_job_config(cls) -> "bigquery.LoadJobConfig":
        return bigquery.LoadJobConfig(
            create_session=False,
            connection_properties=[bigquery.query.ConnectionProperty(key="session_id", value=cls._session_id)],
        )

    @classmethod
    def _get_credentials(cls, source: BigquerySourceSpec, secret_resolver: Optional[SecretResolver]) -> Optional[str]:
        if source.credentials:
            if secret_resolver is None:
                msg = "Missing a secret resolver."
                raise TectonInternalError(msg)
            return secret_resolver.resolve(source.credentials)

    @classmethod
    def get_client(
        cls,
        source: Optional[BigquerySourceSpec] = None,
        secret_resolver: Optional[SecretResolver] = None,
    ) -> "bigquery.Client":
        user_agent = client_info.ClientInfo(user_agent=cls._user_agent)
        if source:
            credentials = cls._get_credentials(source, secret_resolver) if source.credentials else None
            location = source.location
            if credentials:
                client = bigquery.Client.from_service_account_info(
                    json.loads(credentials), location=location, client_info=user_agent
                )
            else:
                client = bigquery.Client(location=location, client_info=user_agent)
        else:
            if cls._session_client:
                return cls._session_client
            client = bigquery.Client(client_info=user_agent)

        if not cls._session_id:
            cls._session_client = client
            cls._initiate_session()

        return client



import datetime
import functools
import itertools
import typing
from typing import List

import pendulum

from tecton_core import specs
from tecton_core.fco_container import DataProto
from tecton_core.fco_container import create_fco_container
from tecton_core.feature_definition_wrapper import FeatureDefinitionWrapper
from tecton_proto.materialization.params__client_pb2 import MaterializationTaskParams


@functools.lru_cache(maxsize=0)
def feature_definition_from_task_params(params: MaterializationTaskParams) -> FeatureDefinitionWrapper:
    fco_container = create_fco_container(
        itertools.chain(
            *(
                # The cast helps out various type inference implementations which otherwise would complain that
                # the different FCO lists are not hte same type
                typing.cast(typing.Iterable[DataProto], p)
                for p in (params.virtual_data_sources, params.transformations, params.entities)
            )
        ),
        deserialize_funcs_to_main=True,
    )
    fv_spec = specs.create_feature_view_spec_from_data_proto(params.feature_view)
    return FeatureDefinitionWrapper(fv_spec, fco_container)


class TimeInterval(typing.NamedTuple):
    start: datetime.datetime
    end: datetime.datetime

    def to_pendulum(self) -> pendulum.Period:
        return pendulum.instance(self.end) - pendulum.instance(self.start)


def _backfill_job_periods(
    start_time: datetime.datetime, end_time: datetime.datetime, interval: datetime.timedelta
) -> List[TimeInterval]:
    jobs = []
    while start_time < end_time:
        jobs.append(TimeInterval(start_time, start_time + interval))
        start_time = start_time + interval
    assert start_time == end_time, "Start and end times were not aligned to `interval`"
    return jobs


def job_query_intervals(task_params: MaterializationTaskParams) -> typing.List[TimeInterval]:
    """
    Return a list of start/end tuples of size batch_schedule.
    For use of breaking up a large backfill window into incremental sizes.
    """
    fd = feature_definition_from_task_params(task_params)
    assert task_params.batch_task_info.HasField("batch_parameters")
    task_feature_start_time = task_params.batch_task_info.batch_parameters.feature_start_time.ToDatetime()
    task_feature_end_time = task_params.batch_task_info.batch_parameters.feature_end_time.ToDatetime()
    # for incremental backfills, we split the job into each batch interval
    if fd.is_incremental_backfill:
        return _backfill_job_periods(task_feature_start_time, task_feature_end_time, fd.batch_materialization_schedule)
    else:
        return [TimeInterval(task_feature_start_time, task_feature_end_time)]



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import sys

from pyspark import since, keyword_only
from pyspark.ml.param.shared import HasPredictionCol, HasBlockSize, HasMaxIter, HasRegParam, \
    HasCheckpointInterval, HasSeed
from pyspark.ml.wrapper import JavaEstimator, JavaModel
from pyspark.ml.common import inherit_doc
from pyspark.ml.param import Params, TypeConverters, Param
from pyspark.ml.util import JavaMLWritable, JavaMLReadable


__all__ = ['ALS', 'ALSModel']


@inherit_doc
class _ALSModelParams(HasPredictionCol, HasBlockSize):
    """
    Params for :py:class:`ALS` and :py:class:`ALSModel`.

    .. versionadded:: 3.0.0
    """

    userCol = Param(Params._dummy(), "userCol", "column name for user ids. Ids must be within " +
                    "the integer value range.", typeConverter=TypeConverters.toString)
    itemCol = Param(Params._dummy(), "itemCol", "column name for item ids. Ids must be within " +
                    "the integer value range.", typeConverter=TypeConverters.toString)
    coldStartStrategy = Param(Params._dummy(), "coldStartStrategy", "strategy for dealing with " +
                              "unknown or new users/items at prediction time. This may be useful " +
                              "in cross-validation or production scenarios, for handling " +
                              "user/item ids the model has not seen in the training data. " +
                              "Supported values: 'nan', 'drop'.",
                              typeConverter=TypeConverters.toString)

    def __init__(self, *args):
        super(_ALSModelParams, self).__init__(*args)
        self._setDefault(blockSize=4096)

    @since("1.4.0")
    def getUserCol(self):
        """
        Gets the value of userCol or its default value.
        """
        return self.getOrDefault(self.userCol)

    @since("1.4.0")
    def getItemCol(self):
        """
        Gets the value of itemCol or its default value.
        """
        return self.getOrDefault(self.itemCol)

    @since("2.2.0")
    def getColdStartStrategy(self):
        """
        Gets the value of coldStartStrategy or its default value.
        """
        return self.getOrDefault(self.coldStartStrategy)


@inherit_doc
class _ALSParams(_ALSModelParams, HasMaxIter, HasRegParam, HasCheckpointInterval, HasSeed):
    """
    Params for :py:class:`ALS`.

    .. versionadded:: 3.0.0
    """

    rank = Param(Params._dummy(), "rank", "rank of the factorization",
                 typeConverter=TypeConverters.toInt)
    numUserBlocks = Param(Params._dummy(), "numUserBlocks", "number of user blocks",
                          typeConverter=TypeConverters.toInt)
    numItemBlocks = Param(Params._dummy(), "numItemBlocks", "number of item blocks",
                          typeConverter=TypeConverters.toInt)
    implicitPrefs = Param(Params._dummy(), "implicitPrefs", "whether to use implicit preference",
                          typeConverter=TypeConverters.toBoolean)
    alpha = Param(Params._dummy(), "alpha", "alpha for implicit preference",
                  typeConverter=TypeConverters.toFloat)

    ratingCol = Param(Params._dummy(), "ratingCol", "column name for ratings",
                      typeConverter=TypeConverters.toString)
    nonnegative = Param(Params._dummy(), "nonnegative",
                        "whether to use nonnegative constraint for least squares",
                        typeConverter=TypeConverters.toBoolean)
    intermediateStorageLevel = Param(Params._dummy(), "intermediateStorageLevel",
                                     "StorageLevel for intermediate datasets. Cannot be 'NONE'.",
                                     typeConverter=TypeConverters.toString)
    finalStorageLevel = Param(Params._dummy(), "finalStorageLevel",
                              "StorageLevel for ALS model factors.",
                              typeConverter=TypeConverters.toString)

    def __init__(self, *args):
        super(_ALSParams, self).__init__(*args)
        self._setDefault(rank=10, maxIter=10, regParam=0.1, numUserBlocks=10, numItemBlocks=10,
                         implicitPrefs=False, alpha=1.0, userCol="user", itemCol="item",
                         ratingCol="rating", nonnegative=False, checkpointInterval=10,
                         intermediateStorageLevel="MEMORY_AND_DISK",
                         finalStorageLevel="MEMORY_AND_DISK", coldStartStrategy="nan")

    @since("1.4.0")
    def getRank(self):
        """
        Gets the value of rank or its default value.
        """
        return self.getOrDefault(self.rank)

    @since("1.4.0")
    def getNumUserBlocks(self):
        """
        Gets the value of numUserBlocks or its default value.
        """
        return self.getOrDefault(self.numUserBlocks)

    @since("1.4.0")
    def getNumItemBlocks(self):
        """
        Gets the value of numItemBlocks or its default value.
        """
        return self.getOrDefault(self.numItemBlocks)

    @since("1.4.0")
    def getImplicitPrefs(self):
        """
        Gets the value of implicitPrefs or its default value.
        """
        return self.getOrDefault(self.implicitPrefs)

    @since("1.4.0")
    def getAlpha(self):
        """
        Gets the value of alpha or its default value.
        """
        return self.getOrDefault(self.alpha)

    @since("1.4.0")
    def getRatingCol(self):
        """
        Gets the value of ratingCol or its default value.
        """
        return self.getOrDefault(self.ratingCol)

    @since("1.4.0")
    def getNonnegative(self):
        """
        Gets the value of nonnegative or its default value.
        """
        return self.getOrDefault(self.nonnegative)

    @since("2.0.0")
    def getIntermediateStorageLevel(self):
        """
        Gets the value of intermediateStorageLevel or its default value.
        """
        return self.getOrDefault(self.intermediateStorageLevel)

    @since("2.0.0")
    def getFinalStorageLevel(self):
        """
        Gets the value of finalStorageLevel or its default value.
        """
        return self.getOrDefault(self.finalStorageLevel)


@inherit_doc
class ALS(JavaEstimator, _ALSParams, JavaMLWritable, JavaMLReadable):
    """
    Alternating Least Squares (ALS) matrix factorization.

    ALS attempts to estimate the ratings matrix `R` as the product of
    two lower-rank matrices, `X` and `Y`, i.e. `X * Yt = R`. Typically
    these approximations are called 'factor' matrices. The general
    approach is iterative. During each iteration, one of the factor
    matrices is held constant, while the other is solved for using least
    squares. The newly-solved factor matrix is then held constant while
    solving for the other factor matrix.

    This is a blocked implementation of the ALS factorization algorithm
    that groups the two sets of factors (referred to as "users" and
    "products") into blocks and reduces communication by only sending
    one copy of each user vector to each product block on each
    iteration, and only for the product blocks that need that user's
    feature vector. This is achieved by pre-computing some information
    about the ratings matrix to determine the "out-links" of each user
    (which blocks of products it will contribute to) and "in-link"
    information for each product (which of the feature vectors it
    receives from each user block it will depend on). This allows us to
    send only an array of feature vectors between each user block and
    product block, and have the product block find the users' ratings
    and update the products based on these messages.

    For implicit preference data, the algorithm used is based on
    `"Collaborative Filtering for Implicit Feedback Datasets",
    <https://doi.org/10.1109/ICDM.2008.22>`_, adapted for the blocked
    approach used here.

    Essentially instead of finding the low-rank approximations to the
    rating matrix `R`, this finds the approximations for a preference
    matrix `P` where the elements of `P` are 1 if r > 0 and 0 if r <= 0.
    The ratings then act as 'confidence' values related to strength of
    indicated user preferences rather than explicit ratings given to
    items.

    .. versionadded:: 1.4.0

    Notes
    -----
    The input rating dataframe to the ALS implementation should be deterministic.
    Nondeterministic data can cause failure during fitting ALS model.
    For example, an order-sensitive operation like sampling after a repartition makes
    dataframe output nondeterministic, like `df.repartition(2).sample(False, 0.5, 1618)`.
    Checkpointing sampled dataframe or adding a sort before sampling can help make the
    dataframe deterministic.

    Examples
    --------
    >>> df = spark.createDataFrame(
    ...     [(0, 0, 4.0), (0, 1, 2.0), (1, 1, 3.0), (1, 2, 4.0), (2, 1, 1.0), (2, 2, 5.0)],
    ...     ["user", "item", "rating"])
    >>> als = ALS(rank=10, seed=0)
    >>> als.setMaxIter(5)
    ALS...
    >>> als.getMaxIter()
    5
    >>> als.setRegParam(0.1)
    ALS...
    >>> als.getRegParam()
    0.1
    >>> als.clear(als.regParam)
    >>> model = als.fit(df)
    >>> model.getBlockSize()
    4096
    >>> model.getUserCol()
    'user'
    >>> model.setUserCol("user")
    ALSModel...
    >>> model.getItemCol()
    'item'
    >>> model.setPredictionCol("newPrediction")
    ALS...
    >>> model.rank
    10
    >>> model.userFactors.orderBy("id").collect()
    [Row(id=0, features=[...]), Row(id=1, ...), Row(id=2, ...)]
    >>> test = spark.createDataFrame([(0, 2), (1, 0), (2, 0)], ["user", "item"])
    >>> predictions = sorted(model.transform(test).collect(), key=lambda r: r[0])
    >>> predictions[0]
    Row(user=0, item=2, newPrediction=0.692910...)
    >>> predictions[1]
    Row(user=1, item=0, newPrediction=3.473569...)
    >>> predictions[2]
    Row(user=2, item=0, newPrediction=-0.899198...)
    >>> user_recs = model.recommendForAllUsers(3)
    >>> user_recs.where(user_recs.user == 0)\
        .select("recommendations.item", "recommendations.rating").collect()
    [Row(item=[0, 1, 2], rating=[3.910..., 1.997..., 0.692...])]
    >>> item_recs = model.recommendForAllItems(3)
    >>> item_recs.where(item_recs.item == 2)\
        .select("recommendations.user", "recommendations.rating").collect()
    [Row(user=[2, 1, 0], rating=[4.892..., 3.991..., 0.692...])]
    >>> user_subset = df.where(df.user == 2)
    >>> user_subset_recs = model.recommendForUserSubset(user_subset, 3)
    >>> user_subset_recs.select("recommendations.item", "recommendations.rating").first()
    Row(item=[2, 1, 0], rating=[4.892..., 1.076..., -0.899...])
    >>> item_subset = df.where(df.item == 0)
    >>> item_subset_recs = model.recommendForItemSubset(item_subset, 3)
    >>> item_subset_recs.select("recommendations.user", "recommendations.rating").first()
    Row(user=[0, 1, 2], rating=[3.910..., 3.473..., -0.899...])
    >>> als_path = temp_path + "/als"
    >>> als.save(als_path)
    >>> als2 = ALS.load(als_path)
    >>> als.getMaxIter()
    5
    >>> model_path = temp_path + "/als_model"
    >>> model.save(model_path)
    >>> model2 = ALSModel.load(model_path)
    >>> model.rank == model2.rank
    True
    >>> sorted(model.userFactors.collect()) == sorted(model2.userFactors.collect())
    True
    >>> sorted(model.itemFactors.collect()) == sorted(model2.itemFactors.collect())
    True
    >>> model.transform(test).take(1) == model2.transform(test).take(1)
    True
    """

    @keyword_only
    def __init__(self, *, rank=10, maxIter=10, regParam=0.1, numUserBlocks=10,
                 numItemBlocks=10, implicitPrefs=False, alpha=1.0, userCol="user", itemCol="item",
                 seed=None, ratingCol="rating", nonnegative=False, checkpointInterval=10,
                 intermediateStorageLevel="MEMORY_AND_DISK",
                 finalStorageLevel="MEMORY_AND_DISK", coldStartStrategy="nan", blockSize=4096):
        """
        __init__(self, \\*, rank=10, maxIter=10, regParam=0.1, numUserBlocks=10,
                 numItemBlocks=10, implicitPrefs=False, alpha=1.0, userCol="user", itemCol="item", \
                 seed=None, ratingCol="rating", nonnegative=False, checkpointInterval=10, \
                 intermediateStorageLevel="MEMORY_AND_DISK", \
                 finalStorageLevel="MEMORY_AND_DISK", coldStartStrategy="nan", blockSize=4096)
        """
        super(ALS, self).__init__()
        self._java_obj = self._new_java_obj("org.apache.spark.ml.recommendation.ALS", self.uid)
        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    @keyword_only
    @since("1.4.0")
    def setParams(self, *, rank=10, maxIter=10, regParam=0.1, numUserBlocks=10,
                  numItemBlocks=10, implicitPrefs=False, alpha=1.0, userCol="user", itemCol="item",
                  seed=None, ratingCol="rating", nonnegative=False, checkpointInterval=10,
                  intermediateStorageLevel="MEMORY_AND_DISK",
                  finalStorageLevel="MEMORY_AND_DISK", coldStartStrategy="nan", blockSize=4096):
        """
        setParams(self, \\*, rank=10, maxIter=10, regParam=0.1, numUserBlocks=10, \
                 numItemBlocks=10, implicitPrefs=False, alpha=1.0, userCol="user", itemCol="item", \
                 seed=None, ratingCol="rating", nonnegative=False, checkpointInterval=10, \
                 intermediateStorageLevel="MEMORY_AND_DISK", \
                 finalStorageLevel="MEMORY_AND_DISK", coldStartStrategy="nan", blockSize=4096)
        Sets params for ALS.
        """
        kwargs = self._input_kwargs
        return self._set(**kwargs)

    def _create_model(self, java_model):
        return ALSModel(java_model)

    @since("1.4.0")
    def setRank(self, value):
        """
        Sets the value of :py:attr:`rank`.
        """
        return self._set(rank=value)

    @since("1.4.0")
    def setNumUserBlocks(self, value):
        """
        Sets the value of :py:attr:`numUserBlocks`.
        """
        return self._set(numUserBlocks=value)

    @since("1.4.0")
    def setNumItemBlocks(self, value):
        """
        Sets the value of :py:attr:`numItemBlocks`.
        """
        return self._set(numItemBlocks=value)

    @since("1.4.0")
    def setNumBlocks(self, value):
        """
        Sets both :py:attr:`numUserBlocks` and :py:attr:`numItemBlocks` to the specific value.
        """
        self._set(numUserBlocks=value)
        return self._set(numItemBlocks=value)

    @since("1.4.0")
    def setImplicitPrefs(self, value):
        """
        Sets the value of :py:attr:`implicitPrefs`.
        """
        return self._set(implicitPrefs=value)

    @since("1.4.0")
    def setAlpha(self, value):
        """
        Sets the value of :py:attr:`alpha`.
        """
        return self._set(alpha=value)

    @since("1.4.0")
    def setUserCol(self, value):
        """
        Sets the value of :py:attr:`userCol`.
        """
        return self._set(userCol=value)

    @since("1.4.0")
    def setItemCol(self, value):
        """
        Sets the value of :py:attr:`itemCol`.
        """
        return self._set(itemCol=value)

    @since("1.4.0")
    def setRatingCol(self, value):
        """
        Sets the value of :py:attr:`ratingCol`.
        """
        return self._set(ratingCol=value)

    @since("1.4.0")
    def setNonnegative(self, value):
        """
        Sets the value of :py:attr:`nonnegative`.
        """
        return self._set(nonnegative=value)

    @since("2.0.0")
    def setIntermediateStorageLevel(self, value):
        """
        Sets the value of :py:attr:`intermediateStorageLevel`.
        """
        return self._set(intermediateStorageLevel=value)

    @since("2.0.0")
    def setFinalStorageLevel(self, value):
        """
        Sets the value of :py:attr:`finalStorageLevel`.
        """
        return self._set(finalStorageLevel=value)

    @since("2.2.0")
    def setColdStartStrategy(self, value):
        """
        Sets the value of :py:attr:`coldStartStrategy`.
        """
        return self._set(coldStartStrategy=value)

    def setMaxIter(self, value):
        """
        Sets the value of :py:attr:`maxIter`.
        """
        return self._set(maxIter=value)

    def setRegParam(self, value):
        """
        Sets the value of :py:attr:`regParam`.
        """
        return self._set(regParam=value)

    def setPredictionCol(self, value):
        """
        Sets the value of :py:attr:`predictionCol`.
        """
        return self._set(predictionCol=value)

    def setCheckpointInterval(self, value):
        """
        Sets the value of :py:attr:`checkpointInterval`.
        """
        return self._set(checkpointInterval=value)

    def setSeed(self, value):
        """
        Sets the value of :py:attr:`seed`.
        """
        return self._set(seed=value)

    @since("3.0.0")
    def setBlockSize(self, value):
        """
        Sets the value of :py:attr:`blockSize`.
        """
        return self._set(blockSize=value)


class ALSModel(JavaModel, _ALSModelParams, JavaMLWritable, JavaMLReadable):
    """
    Model fitted by ALS.

    .. versionadded:: 1.4.0
    """

    @since("3.0.0")
    def setUserCol(self, value):
        """
        Sets the value of :py:attr:`userCol`.
        """
        return self._set(userCol=value)

    @since("3.0.0")
    def setItemCol(self, value):
        """
        Sets the value of :py:attr:`itemCol`.
        """
        return self._set(itemCol=value)

    @since("3.0.0")
    def setColdStartStrategy(self, value):
        """
        Sets the value of :py:attr:`coldStartStrategy`.
        """
        return self._set(coldStartStrategy=value)

    @since("3.0.0")
    def setPredictionCol(self, value):
        """
        Sets the value of :py:attr:`predictionCol`.
        """
        return self._set(predictionCol=value)

    @since("3.0.0")
    def setBlockSize(self, value):
        """
        Sets the value of :py:attr:`blockSize`.
        """
        return self._set(blockSize=value)

    @property
    @since("1.4.0")
    def rank(self):
        """rank of the matrix factorization model"""
        return self._call_java("rank")

    @property
    @since("1.4.0")
    def userFactors(self):
        """
        a DataFrame that stores user factors in two columns: `id` and
        `features`
        """
        return self._call_java("userFactors")

    @property
    @since("1.4.0")
    def itemFactors(self):
        """
        a DataFrame that stores item factors in two columns: `id` and
        `features`
        """
        return self._call_java("itemFactors")

    def recommendForAllUsers(self, numItems):
        """
        Returns top `numItems` items recommended for each user, for all users.

        .. versionadded:: 2.2.0

        Parameters
        ----------
        numItems : int
            max number of recommendations for each user

        Returns
        -------
        :py:class:`pyspark.sql.DataFrame`
            a DataFrame of (userCol, recommendations), where recommendations are
            stored as an array of (itemCol, rating) Rows.
        """
        return self._call_java("recommendForAllUsers", numItems)

    def recommendForAllItems(self, numUsers):
        """
        Returns top `numUsers` users recommended for each item, for all items.

        .. versionadded:: 2.2.0

        Parameters
        ----------
        numUsers : int
            max number of recommendations for each item

        Returns
        -------
        :py:class:`pyspark.sql.DataFrame`
            a DataFrame of (itemCol, recommendations), where recommendations are
            stored as an array of (userCol, rating) Rows.
        """
        return self._call_java("recommendForAllItems", numUsers)

    def recommendForUserSubset(self, dataset, numItems):
        """
        Returns top `numItems` items recommended for each user id in the input data set. Note that
        if there are duplicate ids in the input dataset, only one set of recommendations per unique
        id will be returned.

        .. versionadded:: 2.3.0

        Parameters
        ----------
        dataset : :py:class:`pyspark.sql.DataFrame`
            a DataFrame containing a column of user ids. The column name must match `userCol`.
        numItems : int
            max number of recommendations for each user

        Returns
        -------
        :py:class:`pyspark.sql.DataFrame`
            a DataFrame of (userCol, recommendations), where recommendations are
            stored as an array of (itemCol, rating) Rows.
        """
        return self._call_java("recommendForUserSubset", dataset, numItems)

    def recommendForItemSubset(self, dataset, numUsers):
        """
        Returns top `numUsers` users recommended for each item id in the input data set. Note that
        if there are duplicate ids in the input dataset, only one set of recommendations per unique
        id will be returned.

        .. versionadded:: 2.3.0

        Parameters
        ----------
        dataset : :py:class:`pyspark.sql.DataFrame`
            a DataFrame containing a column of item ids. The column name must match `itemCol`.
        numUsers : int
            max number of recommendations for each item

        Returns
        -------
        :py:class:`pyspark.sql.DataFrame`
            a DataFrame of (itemCol, recommendations), where recommendations are
            stored as an array of (userCol, rating) Rows.
        """
        return self._call_java("recommendForItemSubset", dataset, numUsers)


if __name__ == "__main__":
    import doctest
    import pyspark.ml.recommendation
    from pyspark.sql import SparkSession
    globs = pyspark.ml.recommendation.__dict__.copy()
    # The small batch size here ensures that we see multiple batches,
    # even in these small test examples:
    spark = SparkSession.builder\
        .master("local[2]")\
        .appName("ml.recommendation tests")\
        .getOrCreate()
    sc = spark.sparkContext
    globs['sc'] = sc
    globs['spark'] = spark
    import tempfile
    temp_path = tempfile.mkdtemp()
    globs['temp_path'] = temp_path
    try:
        (failure_count, test_count) = doctest.testmod(globs=globs, optionflags=doctest.ELLIPSIS)
        spark.stop()
    finally:
        from shutil import rmtree
        try:
            rmtree(temp_path)
        except OSError:
            pass
    if failure_count:
        sys.exit(-1)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/common/analytics_options__client.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import descriptor_pb2 as google_dot_protobuf_dot_descriptor__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n3tecton_proto/common/analytics_options__client.proto\x12\x13tecton_proto.common\x1a google/protobuf/descriptor.proto\"V\n\tAnalytics\x12I\n\x0b\x66ilter_type\x18\x01 \x01(\x0e\x32(.tecton_proto.common.AnalyticsFilterTypeR\nfilterType*8\n\x13\x41nalyticsFilterType\x12\x0f\n\x0bNOOP_FILTER\x10\x00\x12\x10\n\x0cPSEUDONYMIZE\x10\x01:\\\n\tanalytics\x12\x1d.google.protobuf.FieldOptions\x18\xae, \x01(\x0b\x32\x1e.tecton_proto.common.AnalyticsR\tanalyticsB\x15\n\x11\x63om.tecton.commonP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.common.analytics_options__client_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:
  google_dot_protobuf_dot_descriptor__pb2.FieldOptions.RegisterExtension(analytics)

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\021com.tecton.commonP\001'
  _ANALYTICSFILTERTYPE._serialized_start=198
  _ANALYTICSFILTERTYPE._serialized_end=254
  _ANALYTICS._serialized_start=110
  _ANALYTICS._serialized_end=196
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/data/batch_data_source__client.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import duration_pb2 as google_dot_protobuf_dot_duration__pb2
from tecton_proto.args import data_source__client_pb2 as tecton__proto_dot_args_dot_data__source____client__pb2
from tecton_proto.args import data_source_config__client_pb2 as tecton__proto_dot_args_dot_data__source__config____client__pb2
from tecton_proto.args import diff_options__client_pb2 as tecton__proto_dot_args_dot_diff__options____client__pb2
from tecton_proto.args import user_defined_function__client_pb2 as tecton__proto_dot_args_dot_user__defined__function____client__pb2
from tecton_proto.common import secret__client_pb2 as tecton__proto_dot_common_dot_secret____client__pb2
from tecton_proto.common import spark_schema__client_pb2 as tecton__proto_dot_common_dot_spark__schema____client__pb2
from tecton_proto.data import hive_metastore__client_pb2 as tecton__proto_dot_data_dot_hive__metastore____client__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n1tecton_proto/data/batch_data_source__client.proto\x12\x11tecton_proto.data\x1a\x1egoogle/protobuf/duration.proto\x1a+tecton_proto/args/data_source__client.proto\x1a\x32tecton_proto/args/data_source_config__client.proto\x1a,tecton_proto/args/diff_options__client.proto\x1a\x35tecton_proto/args/user_defined_function__client.proto\x1a(tecton_proto/common/secret__client.proto\x1a.tecton_proto/common/spark_schema__client.proto\x1a.tecton_proto/data/hive_metastore__client.proto\"T\n\x19TimestampColumnProperties\x12\x1f\n\x0b\x63olumn_name\x18\x01 \x01(\tR\ncolumnName\x12\x16\n\x06\x66ormat\x18\x02 \x01(\tR\x06\x66ormat\"\x9a\x01\n\x1cSparkBatchDataSourceFunction\x12\x42\n\x08\x66unction\x18\x01 \x01(\x0b\x32&.tecton_proto.args.UserDefinedFunctionR\x08\x66unction\x12\x36\n\x17supports_time_filtering\x18\x02 \x01(\x08R\x15supportsTimeFiltering\"\x9b\x01\n\x1dPandasBatchDataSourceFunction\x12\x42\n\x08\x66unction\x18\x01 \x01(\x0b\x32&.tecton_proto.args.UserDefinedFunctionR\x08\x66unction\x12\x36\n\x17supports_time_filtering\x18\x02 \x01(\x08R\x15supportsTimeFiltering\"\xdb\x0b\n\x0f\x42\x61tchDataSource\x12G\n\nhive_table\x18\x01 \x01(\x0b\x32&.tecton_proto.data.HiveTableDataSourceH\x00R\thiveTable\x12\x37\n\x04\x66ile\x18\x08 \x01(\x0b\x32!.tecton_proto.data.FileDataSourceH\x00R\x04\x66ile\x12H\n\x0bredshift_db\x18\x0b \x01(\x0b\x32%.tecton_proto.data.RedshiftDataSourceH\x00R\nredshiftDb\x12\x46\n\tsnowflake\x18\x0c \x01(\x0b\x32&.tecton_proto.data.SnowflakeDataSourceH\x00R\tsnowflake\x12n\n\x1aspark_data_source_function\x18\r \x01(\x0b\x32/.tecton_proto.data.SparkBatchDataSourceFunctionH\x00R\x17sparkDataSourceFunction\x12J\n\x0bunity_table\x18\x0f \x01(\x0b\x32\'.tecton_proto.data.UnityTableDataSourceH\x00R\nunityTable\x12O\n\x11push_source_table\x18\x10 \x01(\x0b\x32!.tecton_proto.data.PushDataSourceH\x00R\x0fpushSourceTable\x12q\n\x1bpandas_data_source_function\x18\x11 \x01(\x0b\x32\x30.tecton_proto.data.PandasBatchDataSourceFunctionH\x00R\x18pandasDataSourceFunction\x12\x43\n\x08\x62igquery\x18\x14 \x01(\x0b\x32%.tecton_proto.data.BigqueryDataSourceH\x00R\x08\x62igquery\x12\x43\n\x0cspark_schema\x18\t \x01(\x0b\x32 .tecton_proto.common.SparkSchemaR\x0bsparkSchema\x12l\n\x1btimestamp_column_properties\x18\x04 \x01(\x0b\x32,.tecton_proto.data.TimestampColumnPropertiesR\x19timestampColumnProperties\x12\x41\n\x0c\x62\x61tch_config\x18\x05 \x01(\x0b\x32\x1e.tecton_proto.args.BatchConfigR\x0b\x62\x61tchConfig\x12\x32\n\x15\x64\x61te_partition_column\x18\x06 \x01(\tR\x13\x64\x61tePartitionColumn\x12h\n\x1a\x64\x61tetime_partition_columns\x18\x07 \x03(\x0b\x32*.tecton_proto.data.DatetimePartitionColumnR\x18\x64\x61tetimePartitionColumns\x12X\n\x14raw_batch_translator\x18\n \x01(\x0b\x32&.tecton_proto.args.UserDefinedFunctionR\x12rawBatchTranslator\x12\x38\n\ndata_delay\x18\x0e \x01(\x0b\x32\x19.google.protobuf.DurationR\tdataDelay\x12I\n\x07secrets\x18\x13 \x03(\x0b\x32/.tecton_proto.data.BatchDataSource.SecretsEntryR\x07secrets\x1a`\n\x0cSecretsEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12:\n\x05value\x18\x02 \x01(\x0b\x32$.tecton_proto.common.SecretReferenceR\x05value:\x02\x38\x01\x42\x0e\n\x0c\x62\x61tch_sourceJ\x04\x08\x02\x10\x03J\x04\x08\x12\x10\x13\"\xaa\x01\n\x14UnityTableDataSource\x12\x18\n\x07\x63\x61talog\x18\x01 \x01(\tR\x07\x63\x61talog\x12\x16\n\x06schema\x18\x02 \x01(\tR\x06schema\x12\x14\n\x05table\x18\x03 \x01(\tR\x05table\x12J\n\x0b\x61\x63\x63\x65ss_mode\x18\x04 \x01(\x0e\x32).tecton_proto.args.UnityCatalogAccessModeR\naccessMode\"\x88\x02\n\x0e\x46ileDataSource\x12\x10\n\x03uri\x18\x01 \x02(\tR\x03uri\x12?\n\x06\x66ormat\x18\x02 \x02(\x0e\x32\'.tecton_proto.data.FileDataSourceFormatR\x06\x66ormat\x12\x33\n\x16\x63onvert_to_glue_format\x18\x04 \x01(\x08R\x13\x63onvertToGlueFormat\x12\x1d\n\nschema_uri\x18\x05 \x01(\tR\tschemaUri\x12I\n\x0fschema_override\x18\x06 \x01(\x0b\x32 .tecton_proto.common.SparkSchemaR\x0eschemaOverrideJ\x04\x08\x03\x10\x04\"\x88\x01\n\x17\x44\x61tetimePartitionColumn\x12\x1f\n\x0b\x63olumn_name\x18\x01 \x02(\tR\ncolumnName\x12#\n\rformat_string\x18\x02 \x02(\tR\x0c\x66ormatString\x12\'\n\x0fminimum_seconds\x18\x03 \x02(\x03R\x0eminimumSeconds\"\xc5\x01\n\x12RedshiftDataSource\x12\x1a\n\x08\x65ndpoint\x18\x01 \x01(\tR\x08\x65ndpoint\x12\x1d\n\ncluster_id\x18\x02 \x01(\tR\tclusterId\x12\x1a\n\x08\x64\x61tabase\x18\x03 \x01(\tR\x08\x64\x61tabase\x12\x16\n\x05table\x18\x04 \x01(\tH\x00R\x05table\x12\x1d\n\x05query\x18\x06 \x01(\tB\x05\x92M\x02\x18\x03H\x00R\x05query\x12\x17\n\x07temp_s3\x18\x05 \x01(\tR\x06tempS3B\x08\n\x06source\"g\n\x13SnowflakeDataSource\x12P\n\rsnowflakeArgs\x18\x01 \x01(\x0b\x32*.tecton_proto.args.SnowflakeDataSourceArgsR\rsnowflakeArgs\"F\n\x0ePushDataSource\x12\x34\n\x16ingested_data_location\x18\x01 \x01(\tR\x14ingestedDataLocation\"\xeb\x01\n\x12\x42igqueryDataSource\x12\x1d\n\nproject_id\x18\x01 \x01(\tR\tprojectId\x12\x18\n\x07\x64\x61taset\x18\x02 \x01(\tR\x07\x64\x61taset\x12\x1a\n\x08location\x18\x03 \x01(\tR\x08location\x12\x16\n\x05table\x18\x04 \x01(\tH\x00R\x05table\x12\x16\n\x05query\x18\x05 \x01(\tH\x00R\x05query\x12\x46\n\x0b\x63redentials\x18\x06 \x01(\x0b\x32$.tecton_proto.common.SecretReferenceR\x0b\x63redentialsB\x08\n\x06source*~\n\x14\x46ileDataSourceFormat\x12 \n\x1c\x46ILE_DATA_SOURCE_FORMAT_JSON\x10\x00\x12#\n\x1f\x46ILE_DATA_SOURCE_FORMAT_PARQUET\x10\x01\x12\x1f\n\x1b\x46ILE_DATA_SOURCE_FORMAT_CSV\x10\x02\x42\x13\n\x0f\x63om.tecton.dataP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.data.batch_data_source__client_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\017com.tecton.dataP\001'
  _BATCHDATASOURCE_SECRETSENTRY._options = None
  _BATCHDATASOURCE_SECRETSENTRY._serialized_options = b'8\001'
  _REDSHIFTDATASOURCE.fields_by_name['query']._options = None
  _REDSHIFTDATASOURCE.fields_by_name['query']._serialized_options = b'\222M\002\030\003'
  _FILEDATASOURCEFORMAT._serialized_start=3537
  _FILEDATASOURCEFORMAT._serialized_end=3663
  _TIMESTAMPCOLUMNPROPERTIES._serialized_start=440
  _TIMESTAMPCOLUMNPROPERTIES._serialized_end=524
  _SPARKBATCHDATASOURCEFUNCTION._serialized_start=527
  _SPARKBATCHDATASOURCEFUNCTION._serialized_end=681
  _PANDASBATCHDATASOURCEFUNCTION._serialized_start=684
  _PANDASBATCHDATASOURCEFUNCTION._serialized_end=839
  _BATCHDATASOURCE._serialized_start=842
  _BATCHDATASOURCE._serialized_end=2341
  _BATCHDATASOURCE_SECRETSENTRY._serialized_start=2217
  _BATCHDATASOURCE_SECRETSENTRY._serialized_end=2313
  _UNITYTABLEDATASOURCE._serialized_start=2344
  _UNITYTABLEDATASOURCE._serialized_end=2514
  _FILEDATASOURCE._serialized_start=2517
  _FILEDATASOURCE._serialized_end=2781
  _DATETIMEPARTITIONCOLUMN._serialized_start=2784
  _DATETIMEPARTITIONCOLUMN._serialized_end=2920
  _REDSHIFTDATASOURCE._serialized_start=2923
  _REDSHIFTDATASOURCE._serialized_end=3120
  _SNOWFLAKEDATASOURCE._serialized_start=3122
  _SNOWFLAKEDATASOURCE._serialized_end=3225
  _PUSHDATASOURCE._serialized_start=3227
  _PUSHDATASOURCE._serialized_end=3297
  _BIGQUERYDATASOURCE._serialized_start=3300
  _BIGQUERYDATASOURCE._serialized_end=3535
# @@protoc_insertion_point(module_scope)



import typing
from collections import deque
from typing import Deque
from typing import Dict
from typing import Optional
from typing import Union

import tecton_core.query.dialect
from tecton_core import specs
from tecton_core.query.executor_params import QueryTreeStep
from tecton_core.query.node_interface import NodeRef
from tecton_core.query.node_interface import QueryNode
from tecton_core.query.node_interface import recurse_query_tree
from tecton_core.query.nodes import AsofJoinFullAggNode
from tecton_core.query.nodes import DataSourceScanNode
from tecton_core.query.nodes import FeatureViewPipelineNode
from tecton_core.query.nodes import MultiOdfvPipelineNode
from tecton_core.query.nodes import PartialAggNode
from tecton_core.query.nodes import StagingNode
from tecton_proto.args.transformation__client_pb2 import TransformationMode


def get_staging_nodes(qt_root: NodeRef, query_tree_step: QueryTreeStep) -> Dict[str, QueryNode]:
    """Returns all StagingNodes in the query tree that belong to the specified QueryTreeStep."""
    # Note: can have the same staging node show up multiple times in the query tree. e.g. GHF(start/end time)
    staging_nodes = {}
    queue: Deque = deque()
    queue.append(qt_root)
    while len(queue) > 0:
        curr_node = queue.pop().node
        if isinstance(curr_node, StagingNode) and curr_node.query_tree_step == query_tree_step:
            staging_nodes[curr_node.staging_table_name_unique()] = curr_node
            continue
        for child_node_ref in curr_node.inputs:
            queue.append(child_node_ref)
    return staging_nodes


def get_first_input_node_of_class(
    qt_root: NodeRef, node_class: type, as_node_ref: bool = False
) -> Union[QueryNode, NodeRef, None]:
    """Returns the first node in the query tree that is an instance of the specified class.

    Optionally returns the NodeRef instead of the QueryNode itself.
    """
    if isinstance(qt_root.node, node_class):
        return qt_root if as_node_ref else qt_root.node

    for child in qt_root.node.inputs:
        node = get_first_input_node_of_class(child, node_class, as_node_ref=as_node_ref)
        if node is not None:
            return node
    return None


def get_pipeline_dialect(qt_root: NodeRef) -> Optional[tecton_core.query.dialect.Dialect]:
    """Returns the dialect of the query tree. Raises an error if more than one dialect is present."""
    encountered_dialects = set()

    def _extract_dialect_from_fv(node: FeatureViewPipelineNode) -> None:
        nonlocal encountered_dialects
        transform_specs = node.feature_definition_wrapper.transformations
        for transform_spec in transform_specs:
            mode = transform_spec.transformation_mode
            if mode == TransformationMode.TRANSFORMATION_MODE_PYTHON:
                encountered_dialects.add(tecton_core.query.dialect.Dialect.PANDAS)
            elif mode == TransformationMode.TRANSFORMATION_MODE_SPARK_SQL:
                encountered_dialects.add(tecton_core.query.dialect.Dialect.SPARK)
            elif mode == TransformationMode.TRANSFORMATION_MODE_PYSPARK:
                encountered_dialects.add(tecton_core.query.dialect.Dialect.SPARK)
            elif mode == TransformationMode.TRANSFORMATION_MODE_SNOWPARK:
                # TODO(danny): should probably be a snowpark dialect
                encountered_dialects.add(tecton_core.query.dialect.Dialect.SNOWFLAKE)
            elif mode == TransformationMode.TRANSFORMATION_MODE_SNOWFLAKE_SQL:
                encountered_dialects.add(tecton_core.query.dialect.Dialect.SNOWFLAKE)
            elif mode == TransformationMode.TRANSFORMATION_MODE_PANDAS:
                encountered_dialects.add(tecton_core.query.dialect.Dialect.PANDAS)
            elif mode == TransformationMode.TRANSFORMATION_MODE_BIGQUERY_SQL:
                encountered_dialects.add(tecton_core.query.dialect.Dialect.BIGQUERY)

    recurse_query_tree(
        qt_root,
        lambda node: _extract_dialect_from_fv(node) if isinstance(node, FeatureViewPipelineNode) else None,
    )

    if len(encountered_dialects) > 1:
        msg = f"Rift does not support retrieving non-materialized features from Feature Views of different transformation modes {[dialect.value for dialect in encountered_dialects]}. Please ensure that at most one transformation mode is used by non-materialized Feature Views in your Feature Service."
        raise Exception(msg)

    if len(encountered_dialects) == 0:
        return None
    return encountered_dialects.pop()


def get_data_source_dialect(qt_root: NodeRef) -> Optional[tecton_core.query.dialect.Dialect]:
    """Returns the dialect of the query tree. Raises an error if more than one dialect is present."""
    encountered_dialects = set()

    def _extract_dialect_from_ds(node: DataSourceScanNode) -> None:
        nonlocal encountered_dialects
        batch_source = node.ds.batch_source
        if isinstance(batch_source, specs.PandasBatchSourceSpec):
            encountered_dialects.add(tecton_core.query.dialect.Dialect.PANDAS)
        elif isinstance(batch_source, (specs.FileSourceSpec, specs.PushTableSourceSpec)):
            encountered_dialects.add(tecton_core.query.dialect.Dialect.DUCKDB)
        elif isinstance(batch_source, specs.SnowflakeSourceSpec):
            encountered_dialects.add(tecton_core.query.dialect.Dialect.SNOWFLAKE)
        elif isinstance(batch_source, specs.BigquerySourceSpec):
            encountered_dialects.add(tecton_core.query.dialect.Dialect.BIGQUERY)
        else:
            msg = f"Unexpected data source type encountered: {batch_source.__class__}"
            raise Exception(msg)

    recurse_query_tree(
        qt_root,
        lambda node: _extract_dialect_from_ds(node) if isinstance(node, DataSourceScanNode) else None,
    )

    # TODO(TEC-16876): Support mixing DWH sources and file sources / Pandas data source functions.
    if len(encountered_dialects) > 1:
        msg = f"Rift does not support having multiple types of data sources: {[dialect.value for dialect in encountered_dialects]}."
        raise Exception(msg)

    if len(encountered_dialects) == 0:
        return None
    return encountered_dialects.pop()


def get_batch_data_sources(qt_root: NodeRef, cls: typing.Type[typing.Any]) -> typing.List[typing.Any]:
    """Returns all BatchSourceSpecs in all DataSourceScanNodes in the tree rooted at qt_root.

    Only BatchSourceSpec subclasses of the given cls are returned.
    """

    def _get_batch_data_sources(ref: NodeRef) -> typing.Iterator[typing.Any]:
        node = ref.node
        for node_input in node.inputs:
            yield from _get_batch_data_sources(node_input)
        if isinstance(node, DataSourceScanNode):
            batch = node.ds.batch_source
            if not batch:
                return
            if isinstance(batch, cls):
                yield batch

    return list(_get_batch_data_sources(qt_root))


def tree_contains(tree: NodeRef, node_type: typing.Type[QueryNode]) -> bool:
    """Returns True if the tree contains a NodeRef of the given type, False otherwise."""
    if isinstance(tree.node, node_type):
        return True

    return any(tree_contains(subtree, node_type) for subtree in tree.inputs)


def pipeline_has_odfvs(qt_root: NodeRef) -> bool:
    return get_first_input_node_of_class(qt_root, node_class=MultiOdfvPipelineNode) is not None


def pipeline_has_aggregations(qt_root: NodeRef) -> bool:
    return any(
        (
            get_first_input_node_of_class(qt_root, node_class=AsofJoinFullAggNode),
            get_first_input_node_of_class(qt_root, node_class=PartialAggNode),
        )
    )



class UserDefinedTransformationError(Exception):
    """Exception that represents a user-caused error occurred during Query Tree evaluation.

    The error is usually caused by a mistake in user configuration or code and
    should be recoverable by user.
    This exception is expected to have a more comprehensive context (in message)
    to be as clear as possible.
    We will most probably drop the traceback of this exception to reduce noise.
    """

    can_drop_traceback = True

    def __init__(self, error_msg):
        self.error_msg = error_msg
        self.context_msg = None

    def with_context(self, context_msg):
        self.context_msg = context_msg
        return self

    def __str__(self):
        msg = ""
        if self.context_msg:
            msg += f"During {self.context_msg} "

        msg += "error caused by user configuration has occurred:\n"
        msg += f"{self.error_msg}\n"
        return msg


class SQLCompilationError(UserDefinedTransformationError):
    def __init__(self, error_msg, sql_query):
        super().__init__(error_msg)
        self.sql_query = sql_query

    def __str__(self):
        return super().__str__() + "Query:\n" + self._query_with_line_numbers()

    def _query_with_line_numbers(self):
        lines = self.sql_query.split("\n")
        return "\n".join(f"{line_no + 1:>3} {line}" for line_no, line in enumerate(lines))


class UserCodeError(UserDefinedTransformationError):
    def __str__(self):
        return super().__str__() + str(self.__cause__)



from types import MappingProxyType
from typing import Any
from typing import Mapping

from tecton._internals.tecton_pydantic import StrictModel


class AggregationFunction(StrictModel):
    base_name: str
    resolved_name: str
    # Use Any for the Mapping value instead of Union[int, float] to avoid automatic Pydantic V1 coercion from int to
    # float. Pydantic V1 will attempt to coerce values in the union in order.
    params: Mapping[str, Any]


# Last N aggregation that doesn't allow duplicates.
def last_distinct(n: int) -> AggregationFunction:
    if not isinstance(n, int):
        msg = "The parameter `n` of the last_distinct aggregation must be an integer."
        raise ValueError(msg)
    return AggregationFunction(base_name="lastn", resolved_name=f"last_distinct_{n}", params=MappingProxyType({"n": n}))


# Last N aggregation that allows duplicates.
def last(n: int) -> AggregationFunction:
    if not isinstance(n, int):
        msg = "The parameter `n` of the last aggregation must be an integer."
        raise ValueError(msg)
    return AggregationFunction(
        base_name="last_non_distinct_n", resolved_name=f"last_{n}", params=MappingProxyType({"n": n})
    )


# First N aggregation that doesn't allow duplicates.
def first_distinct(n: int) -> AggregationFunction:
    if not isinstance(n, int):
        msg = "The parameter `n` of the first_distinct aggregation must be an integer."
        raise ValueError(msg)
    return AggregationFunction(
        base_name="first_distinct_n", resolved_name=f"first_distinct_{n}", params=MappingProxyType({"n": n})
    )


# First N aggregation that allows duplicates.
def first(n: int) -> AggregationFunction:
    if not isinstance(n, int):
        msg = "The parameter `n` of the first aggregation must be an integer."
        raise ValueError(msg)
    return AggregationFunction(
        base_name="first_non_distinct_n", resolved_name=f"first_{n}", params=MappingProxyType({"n": n})
    )


def approx_count_distinct(precision: int = 8) -> AggregationFunction:
    if not isinstance(precision, int):
        msg = "The parameter `precision` of the approx_count_distinct aggregation must be an integer."
        raise ValueError(msg)
    return AggregationFunction(
        base_name="approx_count_distinct",
        resolved_name="approx_count_distinct",
        params=MappingProxyType({"precision": precision}),
    )


def approx_percentile(percentile: float, precision: int = 100) -> AggregationFunction:
    if not isinstance(percentile, float):
        msg = "The parameter `percentile` of the approx_percentile aggregation must be a float."
        raise ValueError(msg)
    if not isinstance(precision, int):
        msg = "The parameter `precision` of the approx_percentile aggregation must be an integer."
        raise ValueError(msg)
    return AggregationFunction(
        base_name="approx_percentile",
        resolved_name=f"approx_percentile_p{str(percentile).replace('.', '_')}",
        params=MappingProxyType({"percentile": percentile, "precision": precision}),
    )



from typing import List

from tecton_proto.data.feature_view__client_pb2 import OnlineServingIndex as OnlineServingIndexProto


class OnlineServingIndex:
    """
    Defines join keys that can be used with ``query_features``.
    This class contains set of partial join keys that will be indexed
    and queryable in FeatureServices that include this feature definition.
    For example, for a feature definition that has join keys ["user", "page"],
    if we pass OnlineServingIndex(["user"]), we will be able to invoke
    feature_service.query_features({"user": "user_1"})
    and get back feature vectors for all user-page pairs where user="user_1".
    """

    def __init__(self, join_keys: List[str]) -> None:
        """
        Create a new OnlineServingIndex.

        :param join_keys: The partial join keys that will be indexed in the online feature store
        """
        self.join_keys = join_keys

    @classmethod
    def from_proto(cls, proto: OnlineServingIndexProto) -> "OnlineServingIndex":
        """Instantiates object from the provided proto object."""
        return cls(list(proto.join_keys))

    def to_proto(self):
        """Returns proto object created from the existing object."""
        proto = OnlineServingIndexProto()
        proto.join_keys.extend(self.join_keys)
        return proto

    def __str__(self):
        return str(self.join_keys)

    def __repr__(self):
        return str(self)



import datetime
from typing import Optional

from typeguard import typechecked

from tecton.framework.data_source import DataSource


class FilteredSource:
    """
    FilteredSource is a convenience utility that can be used to pre-filter ``sources`` in Feature View definitions.

    :param source: Data Source that this FilteredSource class wraps.
    :param start_time_offset: FilteredSource will pre-filter the data source to the time range ``[start_time + start_time_offset, end_time)``. Must be zero or negative. Defaults to zero.
    :return: A FilteredSource to pass into a Feature View.

    Filters ``source`` to the time range ``[start_time + start_time_offset, end_time)`` before executing the feature view transformation. When materializing a range, the output of FeatureViews is automatically filtered by Tecton to the range ``[start_time, end_time)``. In most cases, using a FilteredSource is a performance optimization for when the query engine fails to "push down" timestamp filtering.

    If ``source`` has configured :class:`tecton.declarative.DatetimePartitionColumn`, Tecton will automatically apply partition filters, which can significantly improve performance.

    To filter only on the end time, use ``FilteredSource(credit_scores, start_time_offset=datetime.timedelta.min)``.

    .. code-block:: python

        from tecton import batch_feature_view, BatchSource, HiveConfig, materialization_context, DatetimePartitionColumn
        from tecton import FilteredSource

        # Declare a BatchSource. `timestamp_field` is normally optional, but it is required when the source is used
        # with a FilteredSource. `datetime_partition_columns` is also recommended and will help improve time
        # filtering efficiency.
        partition_columns = [
            DatetimePartitionColumn(column_name="partition_0", datepart="year", zero_padded=True),
            DatetimePartitionColumn(column_name="partition_1", datepart="month", zero_padded=True),
            DatetimePartitionColumn(column_name="partition_2", datepart="day", zero_padded=True),
        ]
        credit_scores = BatchSource(
            name='credit_scores_batch',
            batch_config= HiveConfig(
                database='demo_fraud',
                table='credit_scores',
                timestamp_field='timestamp',
                datetime_partition_columns=partition_columns))

        # Declare a batch feature view with `credit_scores` as a source.
        @batch_feature_view(
            sources=[FilteredSource(credit_scores)],
            ...
        )
        def user_credit_score(credit_scores):
            return f'''
                SELECT user_id, timestamp, credit_score
                FROM {credit_scores}
                '''

        # The above feature view query is equivalent to:
        @batch_feature_view(
            sources=[credit_scores],
            ...
        )
        def user_credit_score(credit_scores, context=materialization_context()):
            return f'''
                WITH FILTERED_CREDIT_SCORES AS (
                    SELECT *
                    FROM {credit_scores}
                    WHERE
                        timestamp >= to_timestamp("{context.start_time}")
                        AND timestamp < to_timestamp("{context.end_time}")
                        -- AND additional clauses if the data source has datetime_partition_columns configured.
                )
                SELECT user_id, timestamp, credit_score
                FROM FILTERED_CREDIT_SCORES
                '''
    """

    @typechecked
    def __init__(self, source: DataSource, start_time_offset: Optional[datetime.timedelta] = None):
        self.source = source
        if start_time_offset is None:
            start_time_offset = datetime.timedelta(0)
        self.start_time_offset = start_time_offset



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import json
import os
import time
import uuid

from pyspark import SparkContext, since
from pyspark.ml.common import inherit_doc
from pyspark.sql import SparkSession
from pyspark.util import VersionUtils


def _jvm():
    """
    Returns the JVM view associated with SparkContext. Must be called
    after SparkContext is initialized.
    """
    jvm = SparkContext._jvm
    if jvm:
        return jvm
    else:
        raise AttributeError("Cannot load _jvm from SparkContext. Is SparkContext initialized?")


class Identifiable(object):
    """
    Object with a unique ID.
    """

    def __init__(self):
        #: A unique id for the object.
        self.uid = self._randomUID()

    def __repr__(self):
        return self.uid

    @classmethod
    def _randomUID(cls):
        """
        Generate a unique string id for the object. The default implementation
        concatenates the class name, "_", and 12 random hex chars.
        """
        return str(cls.__name__ + "_" + uuid.uuid4().hex[-12:])


@inherit_doc
class BaseReadWrite(object):
    """
    Base class for MLWriter and MLReader. Stores information about the SparkContext
    and SparkSession.

    .. versionadded:: 2.3.0
    """

    def __init__(self):
        self._sparkSession = None

    def session(self, sparkSession):
        """
        Sets the Spark Session to use for saving/loading.
        """
        self._sparkSession = sparkSession
        return self

    @property
    def sparkSession(self):
        """
        Returns the user-specified Spark Session or the default.
        """
        if self._sparkSession is None:
            self._sparkSession = SparkSession.builder.getOrCreate()
        return self._sparkSession

    @property
    def sc(self):
        """
        Returns the underlying `SparkContext`.
        """
        return self.sparkSession.sparkContext


@inherit_doc
class MLWriter(BaseReadWrite):
    """
    Utility class that can save ML instances.

    .. versionadded:: 2.0.0
    """

    def __init__(self):
        super(MLWriter, self).__init__()
        self.shouldOverwrite = False
        self.optionMap = {}

    def _handleOverwrite(self, path):
        from pyspark.ml.wrapper import JavaWrapper

        _java_obj = JavaWrapper._new_java_obj("org.apache.spark.ml.util.FileSystemOverwrite")
        wrapper = JavaWrapper(_java_obj)
        wrapper._call_java("handleOverwrite", path, True, self.sparkSession._jsparkSession)

    def save(self, path):
        """Save the ML instance to the input path."""
        if self.shouldOverwrite:
            self._handleOverwrite(path)
        self.saveImpl(path)

    def saveImpl(self, path):
        """
        save() handles overwriting and then calls this method.  Subclasses should override this
        method to implement the actual saving of the instance.
        """
        raise NotImplementedError("MLWriter is not yet implemented for type: %s" % type(self))

    def overwrite(self):
        """Overwrites if the output path already exists."""
        self.shouldOverwrite = True
        return self

    def option(self, key, value):
        """
        Adds an option to the underlying MLWriter. See the documentation for the specific model's
        writer for possible options. The option name (key) is case-insensitive.
        """
        self.optionMap[key.lower()] = str(value)
        return self


@inherit_doc
class GeneralMLWriter(MLWriter):
    """
    Utility class that can save ML instances in different formats.

    .. versionadded:: 2.4.0
    """

    def format(self, source):
        """
        Specifies the format of ML export ("pmml", "internal", or the fully qualified class
        name for export).
        """
        self.source = source
        return self


@inherit_doc
class JavaMLWriter(MLWriter):
    """
    (Private) Specialization of :py:class:`MLWriter` for :py:class:`JavaParams` types
    """

    def __init__(self, instance):
        super(JavaMLWriter, self).__init__()
        _java_obj = instance._to_java()
        self._jwrite = _java_obj.write()

    def save(self, path):
        """Save the ML instance to the input path."""
        if not isinstance(path, str):
            raise TypeError("path should be a string, got type %s" % type(path))
        self._jwrite.save(path)

    def overwrite(self):
        """Overwrites if the output path already exists."""
        self._jwrite.overwrite()
        return self

    def option(self, key, value):
        self._jwrite.option(key, value)
        return self

    def session(self, sparkSession):
        """Sets the Spark Session to use for saving."""
        self._jwrite.session(sparkSession._jsparkSession)
        return self


@inherit_doc
class GeneralJavaMLWriter(JavaMLWriter):
    """
    (Private) Specialization of :py:class:`GeneralMLWriter` for :py:class:`JavaParams` types
    """

    def __init__(self, instance):
        super(GeneralJavaMLWriter, self).__init__(instance)

    def format(self, source):
        """
        Specifies the format of ML export ("pmml", "internal", or the fully qualified class
        name for export).
        """
        self._jwrite.format(source)
        return self


@inherit_doc
class MLWritable(object):
    """
    Mixin for ML instances that provide :py:class:`MLWriter`.

    .. versionadded:: 2.0.0
    """

    def write(self):
        """Returns an MLWriter instance for this ML instance."""
        raise NotImplementedError("MLWritable is not yet implemented for type: %r" % type(self))

    def save(self, path):
        """Save this ML instance to the given path, a shortcut of 'write().save(path)'."""
        self.write().save(path)


@inherit_doc
class JavaMLWritable(MLWritable):
    """
    (Private) Mixin for ML instances that provide :py:class:`JavaMLWriter`.
    """

    def write(self):
        """Returns an MLWriter instance for this ML instance."""
        return JavaMLWriter(self)


@inherit_doc
class GeneralJavaMLWritable(JavaMLWritable):
    """
    (Private) Mixin for ML instances that provide :py:class:`GeneralJavaMLWriter`.
    """

    def write(self):
        """Returns an GeneralMLWriter instance for this ML instance."""
        return GeneralJavaMLWriter(self)


@inherit_doc
class MLReader(BaseReadWrite):
    """
    Utility class that can load ML instances.

    .. versionadded:: 2.0.0
    """

    def __init__(self):
        super(MLReader, self).__init__()

    def load(self, path):
        """Load the ML instance from the input path."""
        raise NotImplementedError("MLReader is not yet implemented for type: %s" % type(self))


@inherit_doc
class JavaMLReader(MLReader):
    """
    (Private) Specialization of :py:class:`MLReader` for :py:class:`JavaParams` types
    """

    def __init__(self, clazz):
        super(JavaMLReader, self).__init__()
        self._clazz = clazz
        self._jread = self._load_java_obj(clazz).read()

    def load(self, path):
        """Load the ML instance from the input path."""
        if not isinstance(path, str):
            raise TypeError("path should be a string, got type %s" % type(path))
        java_obj = self._jread.load(path)
        if not hasattr(self._clazz, "_from_java"):
            raise NotImplementedError("This Java ML type cannot be loaded into Python currently: %r"
                                      % self._clazz)
        return self._clazz._from_java(java_obj)

    def session(self, sparkSession):
        """Sets the Spark Session to use for loading."""
        self._jread.session(sparkSession._jsparkSession)
        return self

    @classmethod
    def _java_loader_class(cls, clazz):
        """
        Returns the full class name of the Java ML instance. The default
        implementation replaces "pyspark" by "org.apache.spark" in
        the Python full class name.
        """
        java_package = clazz.__module__.replace("pyspark", "org.apache.spark")
        if clazz.__name__ in ("Pipeline", "PipelineModel"):
            # Remove the last package name "pipeline" for Pipeline and PipelineModel.
            java_package = ".".join(java_package.split(".")[0:-1])
        return java_package + "." + clazz.__name__

    @classmethod
    def _load_java_obj(cls, clazz):
        """Load the peer Java object of the ML instance."""
        java_class = cls._java_loader_class(clazz)
        java_obj = _jvm()
        for name in java_class.split("."):
            java_obj = getattr(java_obj, name)
        return java_obj


@inherit_doc
class MLReadable(object):
    """
    Mixin for instances that provide :py:class:`MLReader`.

    .. versionadded:: 2.0.0
    """

    @classmethod
    def read(cls):
        """Returns an MLReader instance for this class."""
        raise NotImplementedError("MLReadable.read() not implemented for type: %r" % cls)

    @classmethod
    def load(cls, path):
        """Reads an ML instance from the input path, a shortcut of `read().load(path)`."""
        return cls.read().load(path)


@inherit_doc
class JavaMLReadable(MLReadable):
    """
    (Private) Mixin for instances that provide JavaMLReader.
    """

    @classmethod
    def read(cls):
        """Returns an MLReader instance for this class."""
        return JavaMLReader(cls)


@inherit_doc
class DefaultParamsWritable(MLWritable):
    """
    Helper trait for making simple :py:class:`Params` types writable.  If a :py:class:`Params`
    class stores all data as :py:class:`Param` values, then extending this trait will provide
    a default implementation of writing saved instances of the class.
    This only handles simple :py:class:`Param` types; e.g., it will not handle
    :py:class:`Dataset`. See :py:class:`DefaultParamsReadable`, the counterpart to this trait.

    .. versionadded:: 2.3.0
    """

    def write(self):
        """Returns a DefaultParamsWriter instance for this class."""
        from pyspark.ml.param import Params

        if isinstance(self, Params):
            return DefaultParamsWriter(self)
        else:
            raise TypeError("Cannot use DefautParamsWritable with type %s because it does not " +
                            " extend Params.", type(self))


@inherit_doc
class DefaultParamsWriter(MLWriter):
    """
    Specialization of :py:class:`MLWriter` for :py:class:`Params` types

    Class for writing Estimators and Transformers whose parameters are JSON-serializable.

    .. versionadded:: 2.3.0
    """

    def __init__(self, instance):
        super(DefaultParamsWriter, self).__init__()
        self.instance = instance

    def saveImpl(self, path):
        DefaultParamsWriter.saveMetadata(self.instance, path, self.sc)

    @staticmethod
    def extractJsonParams(instance, skipParams):
        paramMap = instance.extractParamMap()
        jsonParams = {param.name: value for param, value in paramMap.items()
                      if param.name not in skipParams}
        return jsonParams

    @staticmethod
    def saveMetadata(instance, path, sc, extraMetadata=None, paramMap=None):
        """
        Saves metadata + Params to: path + "/metadata"

        - class
        - timestamp
        - sparkVersion
        - uid
        - paramMap
        - defaultParamMap (since 2.4.0)
        - (optionally, extra metadata)

        Parameters
        ----------
        extraMetadata : dict, optional
            Extra metadata to be saved at same level as uid, paramMap, etc.
        paramMap : dict, optional
            If given, this is saved in the "paramMap" field.
        """
        metadataPath = os.path.join(path, "metadata")
        metadataJson = DefaultParamsWriter._get_metadata_to_save(instance,
                                                                 sc,
                                                                 extraMetadata,
                                                                 paramMap)
        sc.parallelize([metadataJson], 1).saveAsTextFile(metadataPath)

    @staticmethod
    def _get_metadata_to_save(instance, sc, extraMetadata=None, paramMap=None):
        """
        Helper for :py:meth:`DefaultParamsWriter.saveMetadata` which extracts the JSON to save.
        This is useful for ensemble models which need to save metadata for many sub-models.

        Notes
        -----
        See :py:meth:`DefaultParamsWriter.saveMetadata` for details on what this includes.
        """
        uid = instance.uid
        cls = instance.__module__ + '.' + instance.__class__.__name__

        # User-supplied param values
        params = instance._paramMap
        jsonParams = {}
        if paramMap is not None:
            jsonParams = paramMap
        else:
            for p in params:
                jsonParams[p.name] = params[p]

        # Default param values
        jsonDefaultParams = {}
        for p in instance._defaultParamMap:
            jsonDefaultParams[p.name] = instance._defaultParamMap[p]

        basicMetadata = {"class": cls, "timestamp": int(round(time.time() * 1000)),
                         "sparkVersion": sc.version, "uid": uid, "paramMap": jsonParams,
                         "defaultParamMap": jsonDefaultParams}
        if extraMetadata is not None:
            basicMetadata.update(extraMetadata)
        return json.dumps(basicMetadata, separators=[',',  ':'])


@inherit_doc
class DefaultParamsReadable(MLReadable):
    """
    Helper trait for making simple :py:class:`Params` types readable.
    If a :py:class:`Params` class stores all data as :py:class:`Param` values,
    then extending this trait will provide a default implementation of reading saved
    instances of the class. This only handles simple :py:class:`Param` types;
    e.g., it will not handle :py:class:`Dataset`. See :py:class:`DefaultParamsWritable`,
    the counterpart to this trait.

    .. versionadded:: 2.3.0
    """

    @classmethod
    def read(cls):
        """Returns a DefaultParamsReader instance for this class."""
        return DefaultParamsReader(cls)


@inherit_doc
class DefaultParamsReader(MLReader):
    """
    Specialization of :py:class:`MLReader` for :py:class:`Params` types

    Default :py:class:`MLReader` implementation for transformers and estimators that
    contain basic (json-serializable) params and no data. This will not handle
    more complex params or types with data (e.g., models with coefficients).

    .. versionadded:: 2.3.0
    """

    def __init__(self, cls):
        super(DefaultParamsReader, self).__init__()
        self.cls = cls

    @staticmethod
    def __get_class(clazz):
        """
        Loads Python class from its name.
        """
        parts = clazz.split('.')
        module = ".".join(parts[:-1])
        m = __import__(module)
        for comp in parts[1:]:
            m = getattr(m, comp)
        return m

    def load(self, path):
        metadata = DefaultParamsReader.loadMetadata(path, self.sc)
        py_type = DefaultParamsReader.__get_class(metadata['class'])
        instance = py_type()
        instance._resetUid(metadata['uid'])
        DefaultParamsReader.getAndSetParams(instance, metadata)
        return instance

    @staticmethod
    def loadMetadata(path, sc, expectedClassName=""):
        """
        Load metadata saved using :py:meth:`DefaultParamsWriter.saveMetadata`

        Parameters
        ----------
        path : str
        sc : :py:class:`pyspark.SparkContext`
        expectedClassName : str, optional
            If non empty, this is checked against the loaded metadata.
        """
        metadataPath = os.path.join(path, "metadata")
        metadataStr = sc.textFile(metadataPath, 1).first()
        loadedVals = DefaultParamsReader._parseMetaData(metadataStr, expectedClassName)
        return loadedVals

    @staticmethod
    def _parseMetaData(metadataStr, expectedClassName=""):
        """
        Parse metadata JSON string produced by :py:meth`DefaultParamsWriter._get_metadata_to_save`.
        This is a helper function for :py:meth:`DefaultParamsReader.loadMetadata`.

        Parameters
        ----------
        metadataStr : str
            JSON string of metadata
        expectedClassName : str, optional
            If non empty, this is checked against the loaded metadata.
        """
        metadata = json.loads(metadataStr)
        className = metadata['class']
        if len(expectedClassName) > 0:
            assert className == expectedClassName, "Error loading metadata: Expected " + \
                "class name {} but found class name {}".format(expectedClassName, className)
        return metadata

    @staticmethod
    def getAndSetParams(instance, metadata, skipParams=None):
        """
        Extract Params from metadata, and set them in the instance.
        """
        # Set user-supplied param values
        for paramName in metadata['paramMap']:
            param = instance.getParam(paramName)
            if skipParams is None or paramName not in skipParams:
                paramValue = metadata['paramMap'][paramName]
                instance.set(param, paramValue)

        # Set default param values
        majorAndMinorVersions = VersionUtils.majorMinorVersion(metadata['sparkVersion'])
        major = majorAndMinorVersions[0]
        minor = majorAndMinorVersions[1]

        # For metadata file prior to Spark 2.4, there is no default section.
        if major > 2 or (major == 2 and minor >= 4):
            assert 'defaultParamMap' in metadata, "Error loading metadata: Expected " + \
                "`defaultParamMap` section not found"

            for paramName in metadata['defaultParamMap']:
                paramValue = metadata['defaultParamMap'][paramName]
                instance._setDefault(**{paramName: paramValue})

    @staticmethod
    def isPythonParamsInstance(metadata):
        return metadata['class'].startswith('pyspark.ml.')

    @staticmethod
    def loadParamsInstance(path, sc):
        """
        Load a :py:class:`Params` instance from the given path, and return it.
        This assumes the instance inherits from :py:class:`MLReadable`.
        """
        metadata = DefaultParamsReader.loadMetadata(path, sc)
        if DefaultParamsReader.isPythonParamsInstance(metadata):
            pythonClassName = metadata['class']
        else:
            pythonClassName = metadata['class'].replace("org.apache.spark", "pyspark")
        py_type = DefaultParamsReader.__get_class(pythonClassName)
        instance = py_type.load(path)
        return instance


@inherit_doc
class HasTrainingSummary(object):
    """
    Base class for models that provides Training summary.

    .. versionadded:: 3.0.0
    """

    @property
    @since("2.1.0")
    def hasSummary(self):
        """
        Indicates whether a training summary exists for this model
        instance.
        """
        return self._call_java("hasSummary")

    @property
    @since("2.1.0")
    def summary(self):
        """
        Gets summary of the model trained on the training set. An exception is thrown if
        no summary exists.
        """
        return (self._call_java("summary"))


class MetaAlgorithmReadWrite:

    @staticmethod
    def isMetaEstimator(pyInstance):
        from pyspark.ml import Estimator, Pipeline
        from pyspark.ml.tuning import _ValidatorParams
        from pyspark.ml.classification import OneVsRest
        return isinstance(pyInstance, Pipeline) or isinstance(pyInstance, OneVsRest) or \
            (isinstance(pyInstance, Estimator) and isinstance(pyInstance, _ValidatorParams))

    @staticmethod
    def getAllNestedStages(pyInstance):
        from pyspark.ml import Pipeline, PipelineModel
        from pyspark.ml.tuning import _ValidatorParams
        from pyspark.ml.classification import OneVsRest, OneVsRestModel

        # TODO: We need to handle `RFormulaModel.pipelineModel` here after Pyspark RFormulaModel
        #  support pipelineModel property.
        if isinstance(pyInstance, Pipeline):
            pySubStages = pyInstance.getStages()
        elif isinstance(pyInstance, PipelineModel):
            pySubStages = pyInstance.stages
        elif isinstance(pyInstance, _ValidatorParams):
            raise ValueError('PySpark does not support nested validator.')
        elif isinstance(pyInstance, OneVsRest):
            pySubStages = [pyInstance.getClassifier()]
        elif isinstance(pyInstance, OneVsRestModel):
            pySubStages = [pyInstance.getClassifier()] + pyInstance.models
        else:
            pySubStages = []

        nestedStages = []
        for pySubStage in pySubStages:
            nestedStages.extend(MetaAlgorithmReadWrite.getAllNestedStages(pySubStage))

        return [pyInstance] + nestedStages

    @staticmethod
    def getUidMap(instance):
        nestedStages = MetaAlgorithmReadWrite.getAllNestedStages(instance)
        uidMap = {stage.uid: stage for stage in nestedStages}
        if len(nestedStages) != len(uidMap):
            raise RuntimeError(f'{instance.__class__.__module__}.{instance.__class__.__name__}'
                               f'.load found a compound estimator with stages with duplicate '
                               f'UIDs. List of UIDs: {list(uidMap.keys())}.')
        return uidMap



from tecton.cli import printer
from tecton_proto.metadataservice import metadata_service__client_pb2 as metadata_service_pb2


class PlanRenderingClient:
    def __init__(
        self,
        response: metadata_service_pb2.QueryStateUpdateResponseV2,
    ):
        self.response = response

    def has_diffs(self) -> bool:
        """Returns True if there are any changes to be applied."""
        return self.num_fcos_changed > 0

    def print_plan(self):
        """Print the tecton plan output to the console."""
        printer.safe_print(self.response.successful_plan_output.string_output)

    def print_empty_plan(self):
        """Print the no changes (empty plan) message to the console."""
        printer.safe_print("🎉 The remote and local state are the same, nothing to do!")

    def print_apply_warnings(self):
        """If tecton apply is run, print relevant plan-level warnings to console."""
        printer.safe_print(self.response.successful_plan_output.apply_warnings)

    def get_json_plan_output(self) -> str:
        """Return plan output as json string."""
        return self.response.successful_plan_output.json_output

    @property
    def num_fcos_changed(self) -> int:
        """Returns the number of FCOs being created, updated, or deleted."""
        return self.response.successful_plan_output.num_fcos_changed



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import cProfile
import pstats
import os
import atexit
import sys

from pyspark.accumulators import AccumulatorParam


class ProfilerCollector(object):
    """
    This class keeps track of different profilers on a per
    stage basis. Also this is used to create new profilers for
    the different stages.
    """

    def __init__(self, profiler_cls, dump_path=None):
        self.profiler_cls = profiler_cls
        self.profile_dump_path = dump_path
        self.profilers = []

    def new_profiler(self, ctx):
        """ Create a new profiler using class `profiler_cls` """
        return self.profiler_cls(ctx)

    def add_profiler(self, id, profiler):
        """ Add a profiler for RDD `id` """
        if not self.profilers:
            if self.profile_dump_path:
                atexit.register(self.dump_profiles, self.profile_dump_path)
            else:
                atexit.register(self.show_profiles)

        self.profilers.append([id, profiler, False])

    def dump_profiles(self, path):
        """ Dump the profile stats into directory `path` """
        for id, profiler, _ in self.profilers:
            profiler.dump(id, path)
        self.profilers = []

    def show_profiles(self):
        """ Print the profile stats to stdout """
        for i, (id, profiler, showed) in enumerate(self.profilers):
            if not showed and profiler:
                profiler.show(id)
                # mark it as showed
                self.profilers[i][2] = True


class Profiler(object):
    """
    PySpark supports custom profilers, this is to allow for different profilers to
    be used as well as outputting to different formats than what is provided in the
    BasicProfiler.

    A custom profiler has to define or inherit the following methods:
        profile - will produce a system profile of some sort.
        stats - return the collected stats.
        dump - dumps the profiles to a path
        add - adds a profile to the existing accumulated profile

    The profiler class is chosen when creating a SparkContext

    Examples
    --------
    >>> from pyspark import SparkConf, SparkContext
    >>> from pyspark import BasicProfiler
    >>> class MyCustomProfiler(BasicProfiler):
    ...     def show(self, id):
    ...         print("My custom profiles for RDD:%s" % id)
    ...
    >>> conf = SparkConf().set("spark.python.profile", "true")
    >>> sc = SparkContext('local', 'test', conf=conf, profiler_cls=MyCustomProfiler)
    >>> sc.parallelize(range(1000)).map(lambda x: 2 * x).take(10)
    [0, 2, 4, 6, 8, 10, 12, 14, 16, 18]
    >>> sc.parallelize(range(1000)).count()
    1000
    >>> sc.show_profiles()
    My custom profiles for RDD:1
    My custom profiles for RDD:3
    >>> sc.stop()

    Notes
    -----
    This API is a developer API.
    """

    def __init__(self, ctx):
        pass

    def profile(self, func):
        """ Do profiling on the function `func`"""
        raise NotImplementedError

    def stats(self):
        """ Return the collected profiling stats (pstats.Stats)"""
        raise NotImplementedError

    def show(self, id):
        """ Print the profile stats to stdout, id is the RDD id """
        stats = self.stats()
        if stats:
            print("=" * 60)
            print("Profile of RDD<id=%d>" % id)
            print("=" * 60)
            stats.sort_stats("time", "cumulative").print_stats()

    def dump(self, id, path):
        """ Dump the profile into path, id is the RDD id """
        if not os.path.exists(path):
            os.makedirs(path)
        stats = self.stats()
        if stats:
            p = os.path.join(path, "rdd_%d.pstats" % id)
            stats.dump_stats(p)


class PStatsParam(AccumulatorParam):
    """PStatsParam is used to merge pstats.Stats"""

    @staticmethod
    def zero(value):
        return None

    @staticmethod
    def addInPlace(value1, value2):
        if value1 is None:
            return value2
        value1.add(value2)
        return value1


class BasicProfiler(Profiler):
    """
    BasicProfiler is the default profiler, which is implemented based on
    cProfile and Accumulator
    """
    def __init__(self, ctx):
        Profiler.__init__(self, ctx)
        # Creates a new accumulator for combining the profiles of different
        # partitions of a stage
        self._accumulator = ctx.accumulator(None, PStatsParam)

    def profile(self, func):
        """ Runs and profiles the method to_profile passed in. A profile object is returned. """
        pr = cProfile.Profile()
        pr.runcall(func)
        st = pstats.Stats(pr)
        st.stream = None  # make it picklable
        st.strip_dirs()

        # Adds a new profile to the existing accumulated value
        self._accumulator.add(st)

    def stats(self):
        return self._accumulator.value


if __name__ == "__main__":
    import doctest
    (failure_count, test_count) = doctest.testmod()
    if failure_count:
        sys.exit(-1)



import argparse
import base64
import json
import logging
import os.path
import re
import sys
from typing import List

from google.protobuf.json_format import MessageToJson
from pyspark.sql import SparkSession

from tecton_materialization.batch_materialization import batch_materialize_from_params
from tecton_materialization.common.job_metadata import LazyJobMetadataClient
from tecton_materialization.delta_maintenance import run_delta_maintenance
from tecton_materialization.entity_deletion import run_offline_store_deleter
from tecton_materialization.entity_deletion import run_online_store_deleter
from tecton_materialization.feature_export import feature_export_from_params
from tecton_materialization.ingest_materialization import ingest_pushed_df
from tecton_materialization.job_metadata import check_spark_job_uniqueness
from tecton_materialization.stream_materialization import stream_materialize_from_params


try:
    import boto3
    from botocore.errorfactory import ClientError
except ImportError as ex:
    print(f"Unable to import boto3 or botocore.errorfactory.ClientError: {ex}")
    # not available and unused in dataproc
    boto3 = None
    ClientError = None

from tecton_core import conf
from tecton_core.id_helper import IdHelper
from tecton_proto.materialization.params__client_pb2 import MaterializationTaskParams


EMR_CLUSTER_INFO_FILE = "job-flow.json"
EMR_CLUSTER_INFO_PATH = f"/mnt/var/lib/info/{EMR_CLUSTER_INFO_FILE}"

logger = logging.getLogger(__name__)


def _deserialize_materialization_task_params(serialized_materialization_task_params) -> MaterializationTaskParams:
    params = MaterializationTaskParams()
    params.ParseFromString(base64.standard_b64decode(serialized_materialization_task_params))
    return params


def _run_id_from_dbutils(dbutils):
    context = dbutils.notebook.entry_point.getDbutils().notebook().getContext().toJson()
    run_id = json.loads(context).get("currentRunId", {}).get("id")
    if not run_id:
        msg = f"Unable to get Databricks run ID from context: {context}"
        raise RuntimeError(msg)
    logger.info(f"Found Databricks run ID: {run_id}")
    return str(run_id)


def _run_id_from_emr():
    try:
        with open(EMR_CLUSTER_INFO_PATH, "r") as f:
            emr_cluster_info = json.load(f)
        run_id = emr_cluster_info["jobFlowId"]
    except Exception as e:
        # for yarn docker runtime the file is mounted (not the entire path)
        try:
            with open(EMR_CLUSTER_INFO_FILE, "r") as f:
                emr_cluster_info = json.load(f)
            run_id = emr_cluster_info["jobFlowId"]
        except Exception:
            logger.error(f"Cluster info on EMR: FAILED with: {e}")
            raise e
    logger.info(f"Found EMR run ID: {run_id}")
    return run_id


def databricks_main(env):
    configure_logging()

    dbutils = env["dbutils"]
    spark = env["spark"]

    conf.set("TECTON_RUNTIME_ENV", "DATABRICKS")
    conf.set("TECTON_RUNTIME_MODE", "MATERIALIZATION")
    conf.set("TECTON_OFFLINE_RETRIEVAL_COMPUTE_MODE", "spark")
    run_id = _run_id_from_dbutils(dbutils)
    serialized_params = dbutils.widgets.get("materialization_params")

    if serialized_params.startswith("s3://"):
        print(f"{serialized_params} appears to be an S3 URI, reading contents")
        bucket, key = _parse_bucket_key_from_uri(serialized_params)
        print(f"Bucket: {bucket}, Key: {key}")
        s3 = boto3.resource("s3")
        params_object = s3.Object(bucket, key)
        serialized_params = params_object.get()["Body"].read()
    elif serialized_params.startswith("dbfs:/"):
        print(f"{serialized_params} appears to be an DBFS Path, reading contents")
        dbfsPath = serialized_params.replace("dbfs:/", "/dbfs/")
        if os.path.exists(dbfsPath):
            with open(dbfsPath, "r") as f:
                serialized_params = f.read().strip()
        else:
            msg = f"Unable to find Materializaton Params in DBFS Path: {dbfsPath}"
            raise RuntimeError(msg)

    params = _deserialize_materialization_task_params(serialized_params)
    main(params, run_id, spark, step=None)


def _redact_fields_from_params(message_json: str, field_names: List[str]):
    for field_name in field_names:
        message_json = re.sub(rf'"{field_name}": ".*?"', f'"{field_name}": "<REDACTED>"', message_json)
    return message_json


def main(params: MaterializationTaskParams, run_id, spark, step, skip_legacy_execution_table_check=False):
    id_ = IdHelper.to_string(params.feature_view.feature_view_id)
    json_params = MessageToJson(params)
    redacted_json_params = _redact_fields_from_params(json_params, ["authenticationToken"])
    msg = f"Starting materialization for the FV '{id_}' params: {redacted_json_params}"
    # Both print and log so it will show up in the log (for Splunk) and on
    # the notebook page
    print(msg)
    logger.info(msg)

    job_metadata_client = LazyJobMetadataClient(params)

    check_spark_job_uniqueness(
        params,
        run_id,
        spark,
        step,
        skip_legacy_execution_table_check,
        job_metadata_client=job_metadata_client,
    )

    # Run job twice if we are injecting a check for idempotency
    tries = 2 if "forced_retry" in params.feature_view.fco_metadata.workspace else 1
    for _ in range(tries):
        if params.HasField("ingest_task_info"):
            assert step is None
            raw_df = spark.read.parquet(params.ingest_task_info.ingest_parameters.ingest_path)
            ingest_pushed_df(spark, raw_df, params)
        elif params.HasField("deletion_task_info"):
            assert step is None
            deletion_parameters = params.deletion_task_info.deletion_parameters
            if deletion_parameters.online:
                run_online_store_deleter(spark, params)
            if deletion_parameters.offline:
                run_offline_store_deleter(spark, params)
        elif params.HasField("delta_maintenance_task_info"):
            assert step is None
            run_delta_maintenance(spark, params)
        elif params.HasField("batch_task_info"):
            batch_materialize_from_params(spark, params, run_id, step=step, job_metadata_client=job_metadata_client)
        elif params.HasField("stream_task_info"):
            stream_materialize_from_params(spark, params, job_metadata_client=job_metadata_client)
        elif params.HasField("feature_export_info"):
            feature_export_from_params(spark, params)
        else:
            msg = "Unknown task info"
            raise Exception(msg)


# There's a separate entrypoint in EMR, since it doesn't have `dbutils` by which we can read input params.
def _parse_bucket_key_from_uri(serialized_params):
    regex = r"s3://(\S+?)/(\S+)"
    match = re.search(regex, serialized_params)
    return match.group(1), match.group(2)


def emr_main() -> None:
    configure_logging()

    parser = argparse.ArgumentParser(description="Tecton materialization library.")
    parser.add_argument(
        "--materialization-params", type=str, help="The parameters for this materialization task", default=None
    )
    parser.add_argument(
        "--spark-session-name", type=str, help="The name of the spark session created for this task", default=None
    )
    parser.add_argument(
        "--materialization-step",
        type=int,
        help="Materialization step",
        default=None,
    )

    parsed_args, unknown_args = parser.parse_known_args()
    print("Parsed: {}".format(parsed_args))
    print("Unknown: {}".format(unknown_args))
    print("Cluster region: {}".format(os.getenv("CLUSTER_REGION")))

    step = parsed_args.materialization_step
    serialized_params = parsed_args.materialization_params
    if os.path.exists(serialized_params):
        print(f"{serialized_params} appears to be a file, reading contents")
        with open(serialized_params, "r") as f:
            serialized_params = f.read()
            print(f"Materialization params contents: {serialized_params}")
    elif serialized_params.startswith("s3://"):
        print(f"{serialized_params} appears to be an S3 URI, reading contents")
        bucket, key = _parse_bucket_key_from_uri(serialized_params)
        print(f"Bucket: {bucket}, Key: {key}")
        s3 = boto3.resource("s3")
        params_object = s3.Object(bucket, key)
        serialized_params = params_object.get()["Body"].read()
    else:
        print(f"{serialized_params} appears to be the contents")

    params = _deserialize_materialization_task_params(serialized_params)

    spark = SparkSession.builder.appName(parsed_args.spark_session_name).getOrCreate()
    run_id = _run_id_from_emr()
    conf.set("TECTON_RUNTIME_ENV", "EMR")
    conf.set("TECTON_RUNTIME_MODE", "MATERIALIZATION")
    conf.set("TECTON_OFFLINE_RETRIEVAL_COMPUTE_MODE", "spark")

    main(params, run_id, spark, step)


def local_test_main(spark, mat_params_path):
    configure_logging()

    with open(mat_params_path, "r") as f:
        serialized_params = f.read()

    params = _deserialize_materialization_task_params(serialized_params)
    conf.set("TECTON_RUNTIME_MODE", "MATERIALIZATION")
    conf.set("TECTON_OFFLINE_RETRIEVAL_COMPUTE_MODE", "spark")
    main(params, 0, spark, step=None)


def configure_logging() -> None:
    # Set the logging level to INFO since materialization logs are internal.
    logging.basicConfig(
        level=logging.INFO,
        stream=sys.stderr,
        format="%(levelname)s - %(asctime)s - %(name)s - %(message)s",
        datefmt="%m/%d/%Y %I:%M:%S %p",
    )

    # This spams logs on INFO or above.
    logging.getLogger("py4j.java_gateway").setLevel(logging.WARN)


if __name__ == "__main__":
    emr_main()



from unittest import TestCase

from tecton._internals.display import Displayable
from tecton_proto.data.summary__client_pb2 import FcoSummary
from tecton_proto.data.summary__client_pb2 import SummaryItem


class DisplayTest(TestCase):
    def test_fco_metadata(self):
        summary_pb = FcoSummary()
        summary_pb.fco_metadata.name = "my_ds"
        summary_pb.fco_metadata.workspace = "dev"
        summary_pb.fco_metadata.created_at.seconds = 1657932800
        summary_pb.fco_metadata.description = "develop"
        summary_pb.fco_metadata.owner = "joe@pizza.com"
        summary_pb.fco_metadata.tags["a"] = "b"
        summary_pb.fco_metadata.source_filename = "foo.py"

        actual = Displayable.from_fco_summary(summary_pb).to_dict()
        expected = {
            "Name": "my_ds",
            "Created At": "2022-07-16 00:53:20 UTC",
            "Workspace": "dev",
            "Description": "develop",
            "Owner": "joe@pizza.com",
            "Tags": {"a": "b"},
            "Last Modified By": "",  # FCO metadata defaults to "" if not specified
            "Source Filename": "foo.py",
        }
        self.assertDictEqual(actual, expected)

    def test_fco_summary(self):
        summary_pb = FcoSummary()

        item = SummaryItem()
        item.display_name = "Some Property"
        item.value = "Some Value"
        summary_pb.summary_items.append(item)

        multi_item = SummaryItem()
        multi_item.display_name = "Some List"
        multi_item.multi_values.extend(["A", "B"])
        summary_pb.summary_items.append(multi_item)

        nested = SummaryItem()
        nested.display_name = "Some Nested Item"
        nested_item1 = SummaryItem()
        nested_item1.display_name = "Some Attribute"
        nested_item1.value = "Value!"
        nested_item2 = SummaryItem()
        nested_item2.display_name = "Some List Attr"
        nested_item2.multi_values.extend(["C", "D"])
        nested_item3 = SummaryItem()
        nested_item3.display_name = "Some Double Nested"
        nested_item3.nested_summary_items.append(nested_item1)
        nested.nested_summary_items.extend([nested_item1, nested_item2, nested_item3])
        summary_pb.summary_items.append(nested)

        actual = Displayable.from_fco_summary(summary_pb).to_dict()
        expected_subset = {
            "Some Property": "Some Value",
            "Some List": ["A", "B"],
            "Some Nested Item": {
                "Some Attribute": "Value!",
                "Some List Attr": ["C", "D"],
                "Some Double Nested": {"Some Attribute": "Value!"},
            },
        }
        self.assertDictContainsSubset(expected_subset, actual)

    def test_display_table(self):
        headings = ["A", "B", "C"]
        rows = [("a1", "b1", "c1"), ("a2", "b2", "c2")]
        actual = Displayable.from_table(headings=headings, rows=rows).to_dict()
        expected = {
            "data": [
                {"A": "a1", "B": "b1", "C": "c1"},
                {"A": "a2", "B": "b2", "C": "c2"},
            ]
        }
        self.assertDictEqual(actual, expected)

        items = [("a1", "b1"), ("a2", "b2")]
        actual = Displayable.from_table(headings=["A", "B"], rows=items).to_dict()
        expected = {
            "data": [
                {
                    "A": "a1",
                    "B": "b1",
                },
                {
                    "A": "a2",
                    "B": "b2",
                },
            ]
        }
        self.assertDictEqual(actual, expected)

    def test_display_properties(self):
        items = [("a1", "b1"), ("a2", "b2")]
        actual = Displayable.from_properties(items=items).to_dict()
        expected = {"a1": "b1", "a2": "b2"}
        self.assertDictEqual(actual, expected)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/args/entity__client.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from tecton_proto.args import basic_info__client_pb2 as tecton__proto_dot_args_dot_basic__info____client__pb2
from tecton_proto.args import diff_options__client_pb2 as tecton__proto_dot_args_dot_diff__options____client__pb2
from tecton_proto.common import framework_version__client_pb2 as tecton__proto_dot_common_dot_framework__version____client__pb2
from tecton_proto.common import id__client_pb2 as tecton__proto_dot_common_dot_id____client__pb2
from tecton_proto.common import schema__client_pb2 as tecton__proto_dot_common_dot_schema____client__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n&tecton_proto/args/entity__client.proto\x12\x11tecton_proto.args\x1a*tecton_proto/args/basic_info__client.proto\x1a,tecton_proto/args/diff_options__client.proto\x1a\x33tecton_proto/common/framework_version__client.proto\x1a$tecton_proto/common/id__client.proto\x1a(tecton_proto/common/schema__client.proto\"\x92\x04\n\nEntityArgs\x12\x34\n\tentity_id\x18\x01 \x01(\x0b\x32\x17.tecton_proto.common.IdR\x08\x65ntityId\x12\x37\n\x04info\x18\x02 \x01(\x0b\x32\x1c.tecton_proto.args.BasicInfoB\x05\x92M\x02\x10\x01R\x04info\x12\x46\n\x07version\x18\x04 \x01(\x0e\x32%.tecton_proto.common.FrameworkVersionB\x05\x92M\x02\x08\x05R\x07version\x12.\n\x0fprevent_destroy\x18\x05 \x01(\x08\x42\x05\x92M\x02\x08\x01R\x0epreventDestroy\x12K\n\x07options\x18\x06 \x03(\x0b\x32*.tecton_proto.args.EntityArgs.OptionsEntryB\x05\x92M\x02\x08\x01R\x07options\x12N\n\x10join_keys_legacy\x18\x03 \x03(\tB$\x92M\r*\x0b\n\tjoin_keys\x92M\x0c*\n\x12\x08\x30.10.0b0\x92M\x02 \x0fR\x0ejoinKeysLegacy\x12\x44\n\tjoin_keys\x18\x07 \x03(\x0b\x32\x1b.tecton_proto.common.ColumnB\n\x92M\x02 \x0f\x92M\x02\x08\x06R\x08joinKeys\x1a:\n\x0cOptionsEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value:\x02\x38\x01\x42\x13\n\x0f\x63om.tecton.argsP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.args.entity__client_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\017com.tecton.argsP\001'
  _ENTITYARGS_OPTIONSENTRY._options = None
  _ENTITYARGS_OPTIONSENTRY._serialized_options = b'8\001'
  _ENTITYARGS.fields_by_name['info']._options = None
  _ENTITYARGS.fields_by_name['info']._serialized_options = b'\222M\002\020\001'
  _ENTITYARGS.fields_by_name['version']._options = None
  _ENTITYARGS.fields_by_name['version']._serialized_options = b'\222M\002\010\005'
  _ENTITYARGS.fields_by_name['prevent_destroy']._options = None
  _ENTITYARGS.fields_by_name['prevent_destroy']._serialized_options = b'\222M\002\010\001'
  _ENTITYARGS.fields_by_name['options']._options = None
  _ENTITYARGS.fields_by_name['options']._serialized_options = b'\222M\002\010\001'
  _ENTITYARGS.fields_by_name['join_keys_legacy']._options = None
  _ENTITYARGS.fields_by_name['join_keys_legacy']._serialized_options = b'\222M\r*\013\n\tjoin_keys\222M\014*\n\022\0100.10.0b0\222M\002 \017'
  _ENTITYARGS.fields_by_name['join_keys']._options = None
  _ENTITYARGS.fields_by_name['join_keys']._serialized_options = b'\222M\002 \017\222M\002\010\006'
  _ENTITYARGS._serialized_start=285
  _ENTITYARGS._serialized_end=815
  _ENTITYARGS_OPTIONSENTRY._serialized_start=757
  _ENTITYARGS_OPTIONSENTRY._serialized_end=815
# @@protoc_insertion_point(module_scope)



import copy
import uuid
from abc import ABC
from abc import abstractmethod
from dataclasses import dataclass
from typing import Any
from typing import Callable
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple

import attrs
import pandas
import pypika
import sqlparse

from tecton_core.compute_mode import ComputeMode
from tecton_core.query.dialect import Dialect
from tecton_core.query.sql_compat import CompatFunctions
from tecton_core.schema import Schema
from tecton_core.vendor.treelib import Tree


@dataclass
class NodeRef:
    """
    Used so we can more easily modify the QueryTree by inserting and removing nodes, e.g.
    def subtree_rewrite(subtree_node_ref):
        subtree_node_ref.node = NewNode(subtree_node_ref.node)
    """

    node: "QueryNode"

    @property
    def columns(self) -> Sequence[str]:
        return self.node.columns

    @property
    def inputs(self) -> Sequence["NodeRef"]:
        return self.node.inputs

    @property
    def input_names(self) -> Optional[List[str]]:
        return self.node.input_names

    @property
    def name(self):
        return self.node.__class__.__name__ + "_" + str(self.node.node_id.hex[:8])

    def as_str(self) -> str:
        return self.node.as_str()

    @property
    def output_schema(self) -> Optional[Schema]:
        return self.node.output_schema

    def pretty_str(
        self,
        node_id: bool = True,
        name: bool = True,
        description: bool = True,
        columns: bool = False,
    ) -> str:
        """Returns a string that represents the query tree which has this NodeRef as the root node.

        Args:
            node_id: If True, the unique id associated with each node will be rendered.
            name: If True, the class names of the nodes will be rendered.
            description: If True, the actions of the nodes will be rendered.
            columns: If True, the columns of each node will be rendered as an appendix after tree itself.
        """
        assert name or description, "At least one of 'name' or 'description' must be True."
        if columns:
            assert node_id, "Can only show columns if 'node_id' is True."
        tree = self.create_tree(node_id=node_id, name=name, description=description)

        # key=False ensures that nodes on the same level are sorted by id and stdout=False ensures a string is returned
        tree_str = tree.show(key=False, stdout=False)

        if not columns:
            return tree_str

        node_columns = []
        max_node_id = tree.size()
        for unique_id in range(1, max_node_id + 1):
            node = tree.get_node(unique_id).data
            node_columns.append(f"<{unique_id}> " + f"{'|'.join(node.columns)}")

        return tree_str + "\n" + "\n".join(node_columns)

    def _to_query(self) -> pypika.queries.QueryBuilder:
        return self.node._to_query()

    def to_sql(self, pretty_sql: bool = False) -> str:
        """
        Attempts to recursively generate sql for this and child nodes.

        Args:
            pretty_sql: If True, the sql will be reformatted and returned as a more readable, multiline string. If False,
            the SQL will be returned as a one line string. For long queries, using pretty_sql=False has better performance.
        """
        return self.node.to_sql(pretty_sql=pretty_sql)

    def create_tree(self, node_id: bool = True, name: bool = True, description: bool = True) -> Tree:
        """Creates a Tree to represent the query tree which has this NodeRef as the root node.

        The Tree is built so that it can immediately generate a string representation.

        Args:
            node_id: If True, the unique id associated with each node will be rendered.
            name: If True, the class names of the nodes will be rendered.
            description: If True, the actions of the nodes will be rendered.
        """
        tree = Tree()
        self._create_tree(tree=tree, parent_id=None, node_id=node_id, name=name, description=description)
        return tree

    def _create_tree(
        self,
        tree: Tree,
        prefix: str = "",
        parent_id: Optional[int] = None,
        node_id: bool = True,
        name: bool = True,
        description: bool = True,
    ) -> None:
        tag = ""

        # Node ids are assigned sequentially, starting with 1.
        unique_id = tree.size() + 1
        if node_id:
            tag += f"<{unique_id}> "

        tag += prefix

        # Add parameter names to the node class. For example, EntityFilterNode(feature_data, entities).
        node_name = self.node.__class__.__name__
        if self.input_names:
            node_name += f"({', '.join(self.input_names)})"

        if name and description:
            tag += f"{node_name}: {self.as_str()}"
        elif name:
            tag += f"{node_name}"
        elif description:
            tag += f"{self.as_str()}"

        # The rendering is messed up if the tag has a newline.
        assert "\n" not in tag

        # We attach this NodeRef so that it can be retrieved later by its node id.
        tree.create_node(tag=tag, identifier=unique_id, parent=parent_id, data=self.node)

        # Recursively handle all children.
        if self.input_names:
            assert len(self.input_names) == len(
                self.inputs
            ), f"`input_names` has length {len(self.input_names)} but `inputs` has length {len(self.inputs)}"
            for input_name, i in zip(self.input_names, self.inputs):
                prefix = f"[{input_name}] "
                i._create_tree(
                    tree=tree,
                    prefix=prefix,
                    parent_id=unique_id,
                    node_id=node_id,
                    name=name,
                    description=description,
                )
        else:
            for i in self.inputs:
                i._create_tree(
                    tree=tree,
                    parent_id=unique_id,
                    node_id=node_id,
                    name=name,
                    description=description,
                )

    def deepcopy(self) -> "NodeRef":
        node_deepcopy = self.node.deepcopy()
        return node_deepcopy.as_ref()


@attrs.frozen
class QueryNode:
    dialect: Dialect = attrs.field()
    compute_mode: ComputeMode = attrs.field()
    func: CompatFunctions = attrs.field(kw_only=True)
    node_id: uuid.UUID = attrs.field(init=False)

    def __attrs_post_init__(self):
        object.__setattr__(self, "node_id", uuid.uuid4())

    @func.default
    def _compat_functions_factory(self):
        return CompatFunctions.for_dialect(self.dialect, self.compute_mode)

    def with_dialect(self, dialect: Dialect, compute_mode: Optional[ComputeMode] = None) -> "QueryNode":
        for node_ref in self.inputs:
            node_ref.node = node_ref.node.with_dialect(dialect, compute_mode)
        return attrs.evolve(self, dialect=dialect, func=CompatFunctions.for_dialect(dialect, compute_mode))

    def deepcopy(self) -> "QueryNode":
        # We avoid using copy.deepcopy as it runs into pickling issues. Instead, we attempt to deepcopy each attribute
        # that requires it.
        node_deepcopy = copy.copy(self)
        field_names = [f.name for f in node_deepcopy.__attrs_attrs__]
        for field_name in field_names:
            field = getattr(node_deepcopy, field_name)
            if isinstance(field, NodeRef):
                child_ref_copy = field.deepcopy()
                changes = {field_name: child_ref_copy}
                node_deepcopy = attrs.evolve(node_deepcopy, **changes)
        return node_deepcopy

    @property
    def columns(self) -> Sequence[str]:
        """
        The columns in the projectlist coming out of this node.
        """
        raise NotImplementedError()

    def as_ref(self) -> NodeRef:
        return NodeRef(self)

    # used for recursing through the tree for tree rewrites
    @property
    def inputs(self) -> Sequence[NodeRef]:
        raise NotImplementedError()

    @property
    def input_names(self) -> Optional[List[str]]:
        """Returns a list of names for the inputs of this node, if all inputs have names. Otherwise returns None.

        If a list is returned, the order of the names should correspond to the order of nodes in the `inputs` property.
        """
        return None

    def as_str(self) -> str:
        """
        Returns a human-readable description of the operation implemented by this node.
        Used by tecton.TectonDataFrame.explain
        """
        raise NotImplementedError()

    def to_sql(self, pretty_sql: bool = False) -> str:
        """
        Attempts to recursively generate sql for this and child nodes.

        Args:
            pretty_sql: If True, the sql will be reformatted and returned as a more readable, multiline string. If False,
            the SQL will be returned as a one line string. For long queries, using pretty_sql=False has better performance.
        """
        sql_str = self._to_query().get_sql()
        if pretty_sql:
            # This can take a very long time for long queries
            return sqlparse.format(sql_str, reindent=True)
        return sql_str

    def get_sql_views(self, pretty_sql: bool = False) -> List[Tuple[str, str]]:
        """
        Get optional sql views for this node. List of Tuple(view_name, view_sql)
        """
        return []

    def _to_query(self) -> pypika.queries.QueryBuilder:
        """
        Attempts to recursively generate sql query for this and child nodes.
        """
        raise NotImplementedError()

    @property
    def output_schema(self) -> Optional[Schema]:
        return


class DataframeWrapper(ABC):
    """
    A wrapper around pyspark, pandas, snowflake, etc dataframes provides a common interface through which we
    can register views.
    """

    @property
    @abstractmethod
    def _dataframe(self) -> Any:  # noqa: ANN401
        """
        The underlying dataframe
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def columns(self) -> List[str]:
        """
        The columns of the dataframe
        """
        raise NotImplementedError

    @property
    def _temp_table_name(self):
        """
        Gets the temp view name registered by register()
        """
        return f"TMP_TABLE_{id(self._dataframe)}"

    def to_pandas(self) -> pandas.DataFrame:
        raise NotImplementedError

    def to_spark(self):
        raise NotImplementedError


def recurse_query_tree(node_ref: NodeRef, f: Callable) -> None:
    f(node_ref.node)
    for child in node_ref.inputs:
        recurse_query_tree(child, f)



from __future__ import annotations

import typing
from abc import abstractmethod
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

import attrs
import pandas
import pendulum
import pypika
from pypika import terms

from tecton_core import feature_definition_wrapper
from tecton_core import query_consts
from tecton_core import time_utils
from tecton_core.compute_mode import ComputeMode
from tecton_core.query import node_interface
from tecton_core.query import nodes
from tecton_core.query.dialect import Dialect
from tecton_core.query.pandas import node as pandas_node
from tecton_core.specs import create_time_window_spec_from_data_proto
from tecton_proto.common import aggregation_function__client_pb2 as afpb
from tecton_proto.data import feature_view__client_pb2 as feature_view_pb2
from tecton_snowflake import pipeline_helper
from tecton_snowflake.query import aggregation_plans
from tecton_snowflake.query import queries


if typing.TYPE_CHECKING:
    import snowflake.snowpark


@attrs.frozen
class PartialAggSnowflakeNode(nodes.PartialAggNode):
    """
    Using a different implementation for LastN/FirstN aggregation functions for Snowflake
    """

    @classmethod
    def from_query_node(cls, query_node: nodes.PartialAggNode) -> "PartialAggSnowflakeNode":
        return cls(
            dialect=Dialect.SNOWFLAKE,
            compute_mode=ComputeMode.SNOWFLAKE,
            input_node=query_node.input_node,
            fdw=query_node.fdw,
            window_start_column_name=query_node.window_start_column_name,
            aggregation_tile_interval=query_node.aggregation_tile_interval,
            window_end_column_name=query_node.window_end_column_name,
            aggregation_anchor_time=query_node.aggregation_anchor_time,
        )

    def _get_partial_agg_columns_and_names(self) -> List[Tuple[terms.Term, str]]:
        """
        Override PartialAggNode's _get_partial_agg_columns to use the Snowflake implementation
        """
        time_aggregation = self.fdw.trailing_time_window_aggregation()
        agg_cols = []
        output_columns = set()
        for feature in time_aggregation.features:
            aggregation_plan = aggregation_plans.get_aggregation_plan(
                feature.function, feature.function_params, self.fdw.time_key
            )
            agg_query_terms = aggregation_plan.partial_aggregation_query_terms(feature.input_feature_name)
            materialized_column_names = aggregation_plan.materialized_column_names(feature.input_feature_name)
            assert len(agg_query_terms) == len(materialized_column_names)
            for column_name, aggregated_column in zip(
                materialized_column_names,
                agg_query_terms,
            ):
                if column_name in output_columns:
                    continue
                output_columns.add(column_name)
                agg_cols.append((aggregated_column, column_name))
        return agg_cols


@attrs.frozen
class AsofJoinFullAggSnowflakeNode(nodes.AsofJoinFullAggNode):
    """
    Asof join full agg rollup, using band joins
    """

    @classmethod
    def from_query_node(cls, query_node: nodes.AsofJoinFullAggNode):
        kwargs = attrs.asdict(query_node, recurse=False)
        del kwargs["node_id"]
        return cls(**kwargs)

    def _to_query(self) -> pypika.Query:
        """
        Snowflake doesn't support RANGE BETWEEN with sliding window as of 07/27/2023, they have it on the roadmap but no ETA.
        This implmentation treat every feature in an aggregation as a separate query, and then join them together.

        :return: Query
        """
        time_aggregation = self.fdw.trailing_time_window_aggregation()
        join_keys = list(self.spine.node.columns)

        # Generate sub queries for each feature
        sub_queries = []
        # Output columns for each sub query
        output_columns = []
        for feature in time_aggregation.features:
            sub_query = self._to_single_feature_aggregate_query(feature)
            sub_queries.append(sub_query)
            output_columns.append([*list(self.spine.node.columns), feature.output_feature_name])

        # Join all sub queries together
        left_columns = output_columns[0]
        result_df = sub_queries[0]
        for right_q, right_columns in zip(sub_queries[1:], output_columns[1:]):
            left_q = result_df
            join_q = queries.SnowflakeQuery().from_(left_q)
            join_q = join_q.inner_join(right_q)
            join_conditions = terms.Criterion.all(
                [
                    terms.Criterion.any(
                        [
                            left_q.field(col) == right_q.field(col),
                            left_q.field(col).isnull() and right_q.field(col).isnull(),
                        ]
                    )
                    for col in join_keys
                ]
            )
            right_nonjoin_cols = set(right_columns) - set(join_keys)
            result_df = join_q.on(join_conditions).select(
                *(left_q.field(col) for col in left_columns), *(right_q.field(col) for col in right_nonjoin_cols)
            )
            left_columns = left_columns + list(right_nonjoin_cols)
        return result_df

    def _to_single_feature_aggregate_query(self, aggregate_feature: feature_view_pb2.AggregateFeature) -> pypika.Query:
        """
        Helper method to run one aggregate feature. This does not use the Range Between implemention.

        :param aggregate_feature: AggregateFeature data proto
        :return: Query
        """
        left_df = self.spine.node._to_query()
        right_df = self.partial_agg_node.node._to_query()
        join_keys = self.fdw.join_keys
        # When calling ghf() with time range, spine only has _ANCHOR_TIME but not time_key
        timestamp_join_cols = (
            [self.fdw.time_key, query_consts.anchor_time()]
            if self.fdw.time_key in self.spine.node.columns
            else [query_consts.anchor_time()]
        )
        common_cols = join_keys + timestamp_join_cols

        output_columns = list(self.spine.node.columns)

        left_name = self.spine.name

        agg_name = self.partial_agg_node.name

        output_feature_name = query_consts.default_case(aggregate_feature.output_feature_name)
        if aggregate_feature.HasField("window"):
            # This is the legacy way to specify an aggregation window
            window_duration = pendulum.Duration(seconds=aggregate_feature.window.ToSeconds())
        else:
            time_window = create_time_window_spec_from_data_proto(aggregate_feature.time_window)
            window_duration = pendulum.Duration(seconds=time_window.window_duration.total_seconds())
        window = time_utils.convert_timedelta_for_version(window_duration, self.fdw.get_feature_store_format_version)
        aggregation_plan = aggregation_plans.get_aggregation_plan(
            aggregate_feature.function,
            aggregate_feature.function_params,
            f"{agg_name}.{query_consts.anchor_time()}",
        )
        names = aggregation_plan.materialized_column_names(aggregate_feature.input_feature_name)
        agg_columns = [*common_cols, aggregation_plan.full_aggregation_join_query_term(names).as_(output_feature_name)]
        # Join spine and partial agg node
        agg_from = (
            queries.SnowflakeQuery()
            .from_(pypika.AliasedQuery(left_name))
            .inner_join(pypika.AliasedQuery(agg_name))
            .on_field(*join_keys)
        )
        # Flatten the input feature if the aggregation function is last/first non-distinct N
        if aggregate_feature.function in {
            afpb.AggregationFunction.AGGREGATION_FUNCTION_LAST_NON_DISTINCT_N,
            afpb.AggregationFunction.AGGREGATION_FUNCTION_FIRST_NON_DISTINCT_N,
        }:
            agg_from = agg_from.lateral(pypika.Field(f"FLATTEN(input=>{names[0]})"))
        # Add the condition that spine time is between partial agg time and partial agg time + window, and run the aggregation function
        agg = (
            agg_from.select(*agg_columns)
            .distinct()
            .where(pypika.AliasedQuery(left_name)._ANCHOR_TIME >= pypika.AliasedQuery(agg_name)._ANCHOR_TIME)
            .where(pypika.AliasedQuery(left_name)._ANCHOR_TIME < pypika.AliasedQuery(agg_name)._ANCHOR_TIME + window)
            .groupby(*(pypika.AliasedQuery(left_name).field(column) for column in output_columns))
        )

        output_feature_name = query_consts.default_case(aggregate_feature.output_feature_name)
        spine_name = left_name + "_" + output_feature_name
        right_name = agg_name + "_AGG"
        feature_column = pypika.AliasedQuery(right_name).field(output_feature_name)
        # TODO(TEC-15924): This behavior is not the same as spark. We should consider consolidating this behavior.
        if aggregate_feature.function == afpb.AggregationFunction.AGGREGATION_FUNCTION_COUNT:
            feature_column = queries.ZeroIfNull(feature_column).as_(output_feature_name)
        # Join the result with the spine node
        join_df = (
            queries.SnowflakeQuery()
            .with_(left_df, spine_name)
            .with_(agg, right_name)
            .from_(pypika.AliasedQuery(spine_name))
            .left_join(pypika.AliasedQuery(right_name))
            .on_field(*common_cols)
            .select(*output_columns, feature_column)
        )
        output_columns.append(output_feature_name)

        return (
            queries.SnowflakeQuery()
            .with_(left_df, left_name)
            .with_(right_df, agg_name)
            .from_(join_df)
            .select(*output_columns)
        )


# SnowparkExecNodes are responsible for taking in a snowpark dataframe, performing some snowpark operation and outputting a
# snowpark dataframe
@attrs.frozen
class SnowparkExecNode:
    columns: List[str]
    input_node: SnowparkExecNode
    column_name_updater: Optional[Callable[[str], str]]  # Snowflake uses this method to uppercase all column names
    session: "snowflake.snowpark.Session"

    @classmethod
    def from_node_inputs(
        cls,
        query_node: node_interface.QueryNode,
        input_node: SnowparkExecNode,
        session: "snowflake.snowpark.Session",
        column_name_updater: Optional[Callable[[str], str]] = lambda x: x,
    ) -> "SnowparkExecNode":
        kwargs = attrs.asdict(query_node, recurse=False)
        kwargs["input_node"] = input_node
        kwargs["columns"] = query_node.columns
        kwargs["column_name_updater"] = column_name_updater
        kwargs["session"] = session
        del kwargs["dialect"]
        del kwargs["compute_mode"]
        del kwargs["func"]
        del kwargs["node_id"]
        return cls(**kwargs)

    def to_dataframe(self) -> pandas.Dataframe:
        return self.to_snowpark().toPandas()

    def to_snowpark(self) -> "snowflake.snowpark.DataFrame":
        df = self._to_snowpark()
        return df

    @abstractmethod
    def _to_snowpark(self) -> "snowflake.snowpark.DataFrame":
        raise NotImplementedError


@attrs.frozen
class SnowparkMultiOdfvPipelineNode(SnowparkExecNode):
    input_node: Union[SnowparkExecNode, pandas_node.SqlExecNode]
    feature_definition_namespaces: List[Tuple[feature_definition_wrapper.FeatureDefinitionWrapper, str]]
    use_namespace_feature_prefix: bool
    events_df_timestamp_field: str

    def _to_snowpark(self) -> "snowflake.snowpark.DataFrame":
        """
        Executes multiple ODFV transformations on the same input dataframe.

        Note: If the user defines their transformation to produce extra columns
        (besides what's specified in output_schema), they will be ignored. If
        there are missing columns they will fail in this function during
        runtime.
        """
        output_df = self.input_node.to_snowpark()

        # Apply ODFV to input df one by one
        for fd, namespace in self.feature_definition_namespaces:
            if fd.is_realtime:
                schema_dict = fd.view_schema.to_dict()
                output_df = pipeline_helper.pipeline_to_df_with_input(
                    session=self.session,
                    input_df=output_df,
                    pipeline=fd.pipeline,
                    transformations=fd.transformations,
                    output_schema=schema_dict,
                    name=fd.name,
                    fv_id=fd.id,
                    namespace=namespace,
                    append_prefix=self.use_namespace_feature_prefix,
                )
        columns_to_drop = [column for column in output_df.columns if query_consts.udf_internal() in column]
        if len(columns_to_drop) > 0:
            output_df = output_df.drop(*columns_to_drop)
        return output_df


@attrs.frozen
class SnowparkRenameColsNode(SnowparkExecNode):
    input_node: Union[SnowparkExecNode, pandas_node.SqlExecNode]
    mapping: Optional[Dict[str, str]]
    drop: Optional[List[str]]

    def _to_snowpark(self) -> "snowflake.snowpark.DataFrame":
        input_df = self.input_node.to_snowpark()
        output_df = input_df
        if self.drop:
            output_df = input_df.drop(*self.drop)
        if self.mapping:
            from snowflake.snowpark.functions import col

            select_columns = []
            for column_name in output_df.columns:
                if column_name not in self.mapping:
                    select_columns.append(col(column_name))
                else:
                    select_columns.append(col(column_name).alias(self.mapping[column_name]))
            output_df = input_df.select(*select_columns)
        return output_df



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/databricks_api/dbfs__client.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n.tecton_proto/databricks_api/dbfs__client.proto\x12\x1btecton_proto.databricks_api\"E\n\x11\x44\x62\x66sDeleteRequest\x12\x12\n\x04path\x18\x01 \x01(\tR\x04path\x12\x1c\n\trecursive\x18\x02 \x01(\x08R\trecursive\"*\n\x14\x44\x62\x66sGetStatusRequest\x12\x12\n\x04path\x18\x01 \x01(\tR\x04path\"U\n\x0f\x44\x62\x66sReadRequest\x12\x12\n\x04path\x18\x01 \x01(\tR\x04path\x12\x16\n\x06offset\x18\x02 \x01(\x03R\x06offset\x12\x16\n\x06length\x18\x03 \x01(\x03R\x06length\"E\n\x10\x44\x62\x66sReadResponse\x12\x1d\n\nbytes_read\x18\x01 \x01(\x03R\tbytesRead\x12\x12\n\x04\x64\x61ta\x18\x02 \x01(\x0cR\x04\x64\x61taB\x1d\n\x19\x63om.tecton.databricks_apiP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.databricks_api.dbfs__client_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\031com.tecton.databricks_apiP\001'
  _DBFSDELETEREQUEST._serialized_start=79
  _DBFSDELETEREQUEST._serialized_end=148
  _DBFSGETSTATUSREQUEST._serialized_start=150
  _DBFSGETSTATUSREQUEST._serialized_end=192
  _DBFSREADREQUEST._serialized_start=194
  _DBFSREADREQUEST._serialized_end=279
  _DBFSREADRESPONSE._serialized_start=281
  _DBFSREADRESPONSE._serialized_end=350
# @@protoc_insertion_point(module_scope)
